import logging
import re
from typing import Mapping

import jsonschema
from jsonschema import Draft7Validator, ValidationError
from jsonschema.validators import extend

from tavern._core import exceptions
from tavern._core.dict_util import recurse_access_key
from tavern._core.exceptions import BadSchemaError
from tavern._core.loader import (
    AnythingSentinel,
    BoolToken,
    FloatToken,
    IntToken,
    RawStrToken,
    TypeConvertToken,
    TypeSentinel,
)
from tavern._core.pytest.config import has_module
from tavern._core.schema.extensions import (
    check_parametrize_marks,
    check_strict_key,
    retry_variable,
    validate_file_spec,
    validate_grpc_status_is_valid_or_list_of_names,
    validate_http_method,
    validate_json_with_ext,
    validate_request_json,
)
from tavern._core.stage_lines import (
    get_stage_filename,
    get_stage_lines,
    read_relevant_lines,
)

logger: logging.Logger = logging.getLogger(__name__)


def is_str_or_bytes_or_token(checker, instance):
    return Draft7Validator.TYPE_CHECKER.is_type(instance, "string") or isinstance(
        instance, (bytes, RawStrToken, AnythingSentinel)
    )


def is_number_or_token(checker, instance):
    return Draft7Validator.TYPE_CHECKER.is_type(instance, "number") or isinstance(
        instance, (IntToken, FloatToken, AnythingSentinel)
    )


def is_integer_or_token(checker, instance):
    return Draft7Validator.TYPE_CHECKER.is_type(instance, "integer") or isinstance(
        instance, (IntToken, AnythingSentinel)
    )


def is_boolean_or_token(checker, instance):
    return Draft7Validator.TYPE_CHECKER.is_type(instance, "boolean") or isinstance(
        instance, (BoolToken, AnythingSentinel)
    )


def is_object_or_sentinel(checker, instance):
    return (
        Draft7Validator.TYPE_CHECKER.is_type(instance, "object")
        or isinstance(instance, (TypeSentinel, TypeConvertToken))
        or instance is None
    )


def oneOf(validator: Draft7Validator, oneOf, instance, schema):
    """Patched version of 'oneof' that does not complain if something is matched by multiple branches"""
    subschemas = enumerate(oneOf)
    all_errors = []
    for index, subschema in subschemas:
        errs = list(validator.descend(instance, subschema, schema_path=index))
        if not errs:
            first_valid = subschema
            break
        all_errors.extend(errs)
    else:
        yield ValidationError(
            f"{instance!r} is not valid under any of the given schemas",
            context=all_errors,
        )

    more_valid = [
        s for i, s in subschemas if validator.evolve(schema=s).is_valid(instance)
    ]
    if more_valid:
        more_valid.append(first_valid)
        reprs = ", ".join(repr(schema) for schema in more_valid)
        logger.debug("%r is valid under each of %s", instance, reprs)


CustomValidator = extend(
    Draft7Validator,
    type_checker=Draft7Validator.TYPE_CHECKER.redefine("object", is_object_or_sentinel)
    .redefine("string", is_str_or_bytes_or_token)
    .redefine("boolean", is_boolean_or_token)
    .redefine("integer", is_integer_or_token)
    .redefine("number", is_number_or_token),
    validators={
        "oneOf": oneOf,
    },
)


def verify_jsonschema(to_verify: Mapping, schema: Mapping) -> None:
    """Verify a generic file against a given jsonschema

    Args:
        to_verify: Filename of source tests to check
        schema: Schema to verify against

    Raises:
        BadSchemaError: Schema did not match
    """

    validator = CustomValidator(schema)

    if "grpc" in to_verify and not has_module("grpc"):
        raise exceptions.BadSchemaError(
            "Tried to use grpc connection string, but grpc was not installed. Reinstall Tavern with the grpc extra like `pip install tavern[grpc]`"
        )
    if "mqtt" in to_verify and not has_module("paho.mqtt"):
        raise exceptions.BadSchemaError(
            "Tried to use mqtt connection string, but mqtt was not installed. Reinstall Tavern with the mqtt extra like `pip install tavern[mqtt]`"
        )

    try:
        validator.validate(to_verify)
    except jsonschema.ValidationError as e:
        real_context = []

        # ignore these strings because they're red herrings
        for c in e.context:
            description = c.schema.get("description", "<no description>")
            if description == "Reference to another stage from an included config file":
                continue

            instance = c.instance
            filename = get_stage_filename(instance)
            if filename is None:
                # Depending on what block raised the error, it mightbe difficult to tell what it was, so check the parent too
                instance = e.instance
                filename = get_stage_filename(instance)

            if filename:
                with open(filename, encoding="utf-8") as infile:
                    n_lines = len(infile.readlines())

                first_line, last_line, _ = get_stage_lines(instance)
                first_line = max(first_line - 2, 0)
                last_line = min(last_line + 2, n_lines)

                reg = re.compile(r"^\s*$")

                lines = read_relevant_lines(instance, first_line, last_line)
                lines = [line for line in lines if not reg.match(line.strip())]
                content = "\n".join(list(lines))
                real_context.append(
                    f"""
{c.message}
{filename}: line {first_line}-{last_line}:

{content}
"""
                )
            else:
                real_context.append(
                    f"""
{c.message}

<error: unable to find input file for context>
"""
                )

        logger.debug("original exception from jsonschema: %s", e)

        msg = "\n---\n" + "\n---\n".join([str(i) for i in real_context])
        raise BadSchemaError(msg) from None

    extra_checks = {
        "stages[*].mqtt_publish.json[]": validate_request_json,
        "stages[*].mqtt_response.payload[]": validate_request_json,
        "stages[*].request.json[]": validate_request_json,
        "stages[*].request.data[]": validate_request_json,
        "stages[*].request.params[]": validate_request_json,
        "stages[*].request.headers[]": validate_request_json,
        "stages[*].grpc_response.status[]": validate_grpc_status_is_valid_or_list_of_names,
        "stages[*].request.method[]": validate_http_method,
        "stages[*].request.save[]": validate_json_with_ext,
        "stages[*].request.files[]": validate_file_spec,
        "marks[*].parametrize[]": check_parametrize_marks,
        "stages[*].response.strict[]": validate_json_with_ext,
        "stages[*].max_retries[]": retry_variable,
        "strict": check_strict_key,
    }

    for path, func in extra_checks.items():
        data = recurse_access_key(to_verify, path)
        if data:
            if path.endswith("[]"):
                if not isinstance(data, list):
                    raise BadSchemaError

                for element in data:
                    func(element, None, path)
            else:
                func(data, None, path)



import functools
import hashlib
import importlib.util
import logging
import os
import string
import subprocess
import sys
import tempfile
from distutils.spawn import find_executable
from importlib.machinery import ModuleSpec
from typing import List

from tavern._core import exceptions

logger: logging.Logger = logging.getLogger(__name__)


@functools.lru_cache
def find_protoc() -> str:
    # Find the Protocol Compiler.
    if "PROTOC" in os.environ and os.path.exists(os.environ["PROTOC"]):
        return os.environ["PROTOC"]

    if protoc := find_executable("protoc"):
        return protoc

    raise exceptions.ProtoCompilerException(
        "Wanted to dynamically compile a proto source, but could not find protoc"
    )


@functools.lru_cache
def _generate_proto_import(source: str) -> None:
    """Invokes the Protocol Compiler to generate a _pb2.py from the given
    .proto file.  Does nothing if the output already exists and is newer than
    the input.
    """

    if not os.path.exists(source):
        raise exceptions.ProtoCompilerException(f"Can't find required file: {source}")

    logger.info("Generating protos from %s...", source)

    # If its a dir, compile them all
    if not os.path.isdir(source):
        if not source.endswith(".proto"):
            raise exceptions.ProtoCompilerException(
                f"invalid proto source file {source}"
            )
        protos = [source]
        include_path = os.path.dirname(source)
    else:
        protos = [
            os.path.join(source, child)
            for child in os.listdir(source)
            if (not os.path.isdir(child)) and child.endswith(".proto")
        ]
        include_path = source

    if not protos:
        raise exceptions.ProtoCompilerException(
            f"No protos defined in {os.path.abspath(source)}"
        )

    for p in protos:
        if not os.path.exists(p):
            raise exceptions.ProtoCompilerException(f"{p} does not exist")

    def sanitise(s):
        """Do basic sanitisation for creating a temporary directory based on
        the name of the input proto file"""
        return "".join(c for c in s if c in string.ascii_letters)

    # Create a temporary directory to put the generated protobuf files in
    output = os.path.join(
        tempfile.gettempdir(),
        "tavern_proto",
        sanitise(protos[0]),
        hashlib.new("sha3_224", "".join(protos).encode("utf8")).hexdigest(),
    )

    if not os.path.exists(output):
        os.makedirs(output)

    protoc = find_protoc()

    protoc_command = [protoc, "-I" + include_path, "--python_out=" + output]
    protoc_command.extend(protos)

    call = subprocess.run(protoc_command, capture_output=True, check=False)  # noqa
    if call.returncode != 0:
        logger.error(f"Error calling '{protoc_command}'")
        raise exceptions.ProtoCompilerException(call.stderr.decode("utf8"))

    logger.info(f"Generated module from protos: {protos}")

    # Invalidate caches so the module can be loaded
    sys.path.append(output)
    importlib.invalidate_caches()
    _import_grpc_module(output)


def _import_grpc_module(python_module_name: str) -> None:
    """takes an expected python module name and tries to import the relevant
    file, adding service to the symbol database.
    """

    logger.debug("attempting to import %s", python_module_name)

    if python_module_name.endswith(".py"):
        raise exceptions.GRPCServiceException(
            f"grpc module definitions should not end with .py, but got {python_module_name}"
        )

    if python_module_name.startswith("."):
        raise exceptions.GRPCServiceException(
            f"relative imports for Python grpc modules not allowed (got {python_module_name})"
        )

    import_specs: List[ModuleSpec] = []

    # Check if its already on the python path
    if (spec := importlib.util.find_spec(python_module_name)) is not None:
        logger.debug(f"{python_module_name} on sys path already")
        import_specs.append(spec)

    # See if the file exists
    module_path = python_module_name.replace(".", "/") + ".py"
    if os.path.exists(module_path):
        logger.debug(f"{python_module_name} found in file")
        if (
            spec := importlib.util.spec_from_file_location(
                python_module_name, module_path
            )
        ) is not None:
            import_specs.append(spec)

    # If its a dir then load files in the dir instead
    if os.path.isdir(python_module_name):
        for s in os.listdir(python_module_name):
            s = os.path.join(python_module_name, s)
            if s.endswith(".py"):
                logger.debug(f"found py file {s}")
                # Guess a package name
                if (
                    spec := importlib.util.spec_from_file_location(s[:-3], s)
                ) is not None:
                    import_specs.append(spec)

    if not import_specs:
        raise exceptions.GRPCServiceException(
            f"could not determine how to import {python_module_name}"
        )

    # Actually import them to register them in the symbol db
    for spec in import_specs:
        mod = importlib.util.module_from_spec(spec)
        logger.debug(f"loading from {spec.name}")
        if spec.loader:
            spec.loader.exec_module(mod)



import logging
from os.path import abspath, dirname, join

import yaml

from tavern._core.dict_util import format_keys
from tavern._core.pytest.config import TestConfig

from .client import GRPCClient
from .request import GRPCRequest
from .response import GRPCResponse

logger: logging.Logger = logging.getLogger(__name__)


session_type = GRPCClient

request_type = GRPCRequest
request_block_name = "grpc_request"


def get_expected_from_request(
    response_block, test_block_config: TestConfig, session: GRPCClient
):
    f_expected = format_keys(response_block, test_block_config.variables)
    expected = f_expected

    return expected


verifier_type = GRPCResponse
response_block_name = "grpc_response"

schema_path: str = join(abspath(dirname(__file__)), "jsonschema.yaml")
with open(schema_path) as schema_file:
    schema = yaml.load(schema_file, Loader=yaml.SafeLoader)



from .hooks import pytest_addhooks, pytest_addoption, pytest_collect_file
from .newhooks import call_hook
from .util import add_parser_options

__all__ = [
    "pytest_addoption",
    "pytest_collect_file",
    "pytest_addhooks",
    "add_parser_options",
    "call_hook",
]



import contextlib
import copy
import os
import tempfile
from collections import OrderedDict
from textwrap import dedent
from unittest.mock import Mock, patch

import pytest
import yaml

from tavern._core import exceptions
from tavern._core.dict_util import (
    check_keys_match_recursive,
    deep_dict_merge,
    format_keys,
    recurse_access_key,
)
from tavern._core.loader import (
    ANYTHING,
    DictSentinel,
    FloatSentinel,
    IncludeLoader,
    IntSentinel,
    ListSentinel,
    StrSentinel,
    construct_include,
    load_single_document_yaml,
)
from tavern._core.schema.extensions import validate_extensions
from tavern._core.schema.files import wrapfile


class TestValidateFunctions:
    def test_get_extension(self):
        """Loads a validation function correctly

        This doesn't check the signature at the time of writing
        """

        spec = {"function": "operator:add"}

        validate_extensions(spec, None, None)

    def test_get_extension_list(self):
        """Loads a validation function correctly

        This doesn't check the signature at the time of writing
        """

        spec = [{"function": "operator:add"}]

        validate_extensions(spec, None, None)

    def test_get_extension_list_empty(self):
        """Loads a validation function correctly

        This doesn't check the signature at the time of writing
        """

        spec = []

        validate_extensions(spec, None, None)

    def test_get_invalid_module(self):
        """Nonexistent module"""

        spec = {"function": "bleuuerhug:add"}

        with pytest.raises(exceptions.BadSchemaError):
            validate_extensions(spec, None, None)

    def test_get_nonexistent_function(self):
        """No name in module"""

        spec = {"function": "os:aaueurhg"}

        with pytest.raises(exceptions.BadSchemaError):
            validate_extensions(spec, None, None)


class TestDictMerge:
    def test_single_level(self):
        """Merge two depth-one dicts with no conflicts"""
        dict_1 = {"key_1": "original_value_1", "key_2": "original_value_2"}
        dict_2 = {"key_2": "new_value_2", "key_3": "new_value_3"}

        result = deep_dict_merge(dict_1, dict_2)

        assert dict_1 == {"key_1": "original_value_1", "key_2": "original_value_2"}
        assert dict_2 == {"key_2": "new_value_2", "key_3": "new_value_3"}
        assert result == {
            "key_1": "original_value_1",
            "key_2": "new_value_2",
            "key_3": "new_value_3",
        }

    def test_recursive_merge(self):
        """Merge two depth-one dicts with no conflicts"""
        dict_1 = {
            "key": {"deep_key_1": "original_value_1", "deep_key_2": "original_value_2"}
        }
        dict_2 = {"key": {"deep_key_2": "new_value_2", "deep_key_3": "new_value_3"}}

        result = deep_dict_merge(dict_1, dict_2)

        assert dict_1 == {
            "key": {"deep_key_1": "original_value_1", "deep_key_2": "original_value_2"}
        }
        assert dict_2 == {
            "key": {"deep_key_2": "new_value_2", "deep_key_3": "new_value_3"}
        }
        assert result == {
            "key": {
                "deep_key_1": "original_value_1",
                "deep_key_2": "new_value_2",
                "deep_key_3": "new_value_3",
            }
        }


class TestMatchRecursive:
    def test_match_dict(self):
        a = {"a": [{"b": "val"}]}
        b = copy.deepcopy(a)

        check_keys_match_recursive(a, b, [])

    def test_match_dict_mismatch(self):
        a = {"a": [{"b": "val"}]}
        b = copy.deepcopy(a)

        b["a"][0]["b"] = "wrong"

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

    def test_match_nested_list(self):
        a = {"a": ["val"]}
        b = copy.deepcopy(a)

        b["a"][0] = "wrong"

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

    def test_match_nested_list_length(self):
        a = {"a": ["val"]}
        b = copy.deepcopy(a)

        b["a"].append("wrong")

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

    # These ones are testing 'internal' behaviour, might break in future

    def test_match_nested_anything_dict(self):
        a = {"a": [{"b": ANYTHING}]}
        b = copy.deepcopy(a)

        b["a"][0]["b"] = "wrong"

        check_keys_match_recursive(a, b, [])

    def test_match_nested_anything_list(self):
        a = {"a": [ANYTHING]}
        b = copy.deepcopy(a)

        b["a"][0] = "wrong"

        check_keys_match_recursive(a, b, [])

    def test_match_ordered(self):
        """Should be able to match an ordereddict"""
        first = {"a": 1, "b": 2}

        second = OrderedDict(b=2, a=1)

        check_keys_match_recursive(first, second, [])

    def test_key_case_matters(self):
        """Make sure case of keys matters"""
        a = {"a": [{"b": "val"}]}
        b = copy.deepcopy(a)
        b["a"][0] = {"B": "val"}

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

    def test_value_case_matters(self):
        """Make sure case of values matters"""
        a = {"a": [{"b": "val"}]}
        b = copy.deepcopy(a)
        b["a"][0]["b"] = "VAL"

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

    @pytest.mark.parametrize(
        "token, response",
        [
            (IntSentinel(), 2),
            (ListSentinel(), [1, 2, 3]),
            (DictSentinel(), {2: 2}),
            (FloatSentinel(), 4.5),
            (StrSentinel(), "dood"),
        ],
    )
    def test_type_token_matches(self, token, response):
        """Make sure type tokens match with generic types"""
        check_keys_match_recursive(token, response, [])

    @pytest.mark.parametrize(
        "token, response",
        [
            (IntSentinel(), 2.3),
            (ListSentinel(), 1),
            (DictSentinel(), [4, 5, 6]),
            (FloatSentinel(), "4"),
            (StrSentinel(), {"a": 2}),
        ],
    )
    def test_type_token_no_match_errors(self, token, response):
        """Make sure type tokens do not match if the type is wrong"""
        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(token, response, [])


class TestNonStrictListMatching:
    def test_match_list_items(self):
        """Should match any 2 list items if strict is False, not if it's True"""
        a = ["b"]
        b = ["a", "b", "c"]

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

        check_keys_match_recursive(a, b, [], strict=False)

    def test_match_multiple(self):
        """As long as they are in the right order, it can match multiple
        items"""
        a = ["a", "c"]
        b = ["a", "b", "c"]

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

        check_keys_match_recursive(a, b, [], strict=False)

    def test_match_multiple_wrong_order(self):
        """Raises an error if the expected items are in the wrong order"""
        a = ["c", "a"]
        b = ["a", "b", "c"]

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [], strict=False)

    def test_match_wrong_type(self):
        """Can't match incorrect type"""
        a = [1]
        b = ["1", "2", "3"]

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [], strict=False)

    def test_match_list_items_more_as(self):
        """One of them is present, the others aren't"""
        a = ["a", "b", "c"]
        b = ["a"]

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [])

        with pytest.raises(exceptions.KeyMismatchError):
            check_keys_match_recursive(a, b, [], strict=False)


@pytest.fixture(name="test_yaml")
def fix_test_yaml():
    text = dedent(
        """
    ---
    test_name: Make sure server doubles number properly

    stages:
      - name: Make sure number is returned correctly
        request:
          url: http://localhost:5000/double
          json:
            is_sensitive: !bool "False"
            raw_str: !raw '{"query": "{ val1 { val2 { val3 { val4, val5 } } } }"}'
            number: !int '5'
            return_float: !bool "True"
          method: POST
          headers:
            content-type: application/json
        response:
          status_code: 200
          json:
            double: !float 10
    """
    )

    return text


class TestCustomTokens:
    def assert_type_value(self, test_value, expected_type, expected_value):
        assert isinstance(test_value, expected_type)
        assert test_value == expected_value

    def test_conversion(self, test_yaml):
        stages = yaml.load(test_yaml, Loader=IncludeLoader)["stages"][0]

        self.assert_type_value(stages["request"]["json"]["number"], int, 5)
        self.assert_type_value(stages["response"]["json"]["double"], float, 10.0)
        self.assert_type_value(stages["request"]["json"]["return_float"], bool, True)
        self.assert_type_value(stages["request"]["json"]["is_sensitive"], bool, False)
        self.assert_type_value(
            stages["request"]["json"]["raw_str"],
            str,
            '{{"query": "{{ val1 {{ val2 {{ val3 {{ val4, val5 }} }} }} }}"}}',
        )


class TestFormatKeys:
    def test_format_missing_raises(self):
        to_format = {"a": "{b}"}

        with pytest.raises(exceptions.MissingFormatError):
            format_keys(to_format, {})

    def test_format_success(self):
        to_format = {"a": "{b}"}

        final_value = "formatted"

        format_variables = {"b": final_value}

        assert format_keys(to_format, format_variables)["a"] == final_value

    def test_no_double_format_failure(self):
        to_format = "{{b}}"

        final_value = "{b}"

        format_variables = {"b": final_value}

        formatted = format_keys(to_format, format_variables)
        assert formatted == final_value
        formatted_2 = format_keys(formatted, {})
        assert formatted_2 == final_value


class TestRecurseAccess:
    @pytest.fixture
    def nested_data(self):
        data = {"a": ["b", {"c": "d"}]}

        return data

    @pytest.mark.parametrize(
        "old_query, new_query, expected_data",
        (("a.0", "a[0]", "b"), ("a.1.c", "a[1].c", "d")),
    )
    def test_search_old_style(self, nested_data, old_query, new_query, expected_data):
        """Make sure old style searches perform the same as jmes queries"""

        with pytest.raises(exceptions.JMESError):
            recurse_access_key(nested_data, old_query)

        new_val = recurse_access_key(nested_data, new_query)
        assert new_val == expected_data

    @pytest.mark.parametrize("new_query", ("f", "a[3]", "a[1].x"))
    def test_missing_search(self, nested_data, new_query):
        """Searching for data not in given data returns None, because of the way the jmespath library works..."""

        assert recurse_access_key(nested_data, new_query) is None


class TestLoadCfg:
    def test_load_one(self):
        example = {"a": "b"}

        with wrapfile(example) as f:
            assert example == load_single_document_yaml(f)

    def test_load_multiple_fails(self):
        example = [{"a": "b"}, {"c": "d"}]

        with tempfile.NamedTemporaryFile(suffix=".yaml", delete=False) as wrapped_tmp:
            # put into a file
            dumped = yaml.dump_all(example)
            wrapped_tmp.write(dumped.encode("utf8"))
            wrapped_tmp.close()

            try:
                with pytest.raises(exceptions.UnexpectedDocumentsError):
                    load_single_document_yaml(wrapped_tmp.name)
            finally:
                os.remove(wrapped_tmp.name)

    @pytest.mark.parametrize("value", ("b", "三木"))
    def test_load_utf8(self, value):
        """if yaml has utf8 char , may load error"""
        content = f"""a: {value}""" ""

        with tempfile.NamedTemporaryFile(suffix=".yaml", delete=False) as f:
            f.write(content.encode("utf8"))

        try:
            load_single_document_yaml(f.name)

        finally:
            os.remove(f.name)


class TestLoadFile:
    @staticmethod
    @contextlib.contextmanager
    def magic_wrap(to_wrap, suffix):
        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as wrapped_tmp:
            dumped = yaml.dump(to_wrap, default_flow_style=False)
            wrapped_tmp.write(dumped.encode("utf8"))
            wrapped_tmp.close()

            try:
                yield wrapped_tmp.name
            finally:
                os.remove(wrapped_tmp.name)

    @pytest.mark.parametrize("suffix", (".yaml", ".yml", ".json"))
    def test_load_extensions(self, suffix):
        example = {"a": "b"}

        with TestLoadFile.magic_wrap(example, suffix) as tmpfile:
            with patch("tavern._core.loader.os.path.join", return_value=tmpfile):
                assert example == construct_include(Mock(), Mock())

    def test_load_bad_extension(self):
        example = {"a": "b"}

        with TestLoadFile.magic_wrap(example, ".bllakjf") as tmpfile:
            with patch("tavern._core.loader.os.path.join", return_value=tmpfile):
                with pytest.raises(exceptions.BadSchemaError):
                    construct_include(Mock(), Mock())

    def test_include_path(self):
        example = {"a": "b"}

        with TestLoadFile.magic_wrap(example, ".yaml") as tmpfile:
            tmppath, tmpfilename = os.path.split(tmpfile)
            with pytest.raises(exceptions.BadSchemaError):
                construct_include(
                    Mock(
                        _root="/does-not-exist", construct_scalar=lambda x: tmpfilename
                    ),
                    Mock(),
                )

            with patch("tavern._core.loader.IncludeLoader.env_path_list", None):
                assert example == construct_include(
                    Mock(_root=tmppath, construct_scalar=lambda x: tmpfilename), Mock()
                )

            os.environ[IncludeLoader.env_var_name] = tmppath
            with patch("tavern._core.loader.IncludeLoader.env_path_list", None):
                assert example == construct_include(
                    Mock(
                        _root="/does-not-exist", construct_scalar=lambda x: tmpfilename
                    ),
                    Mock(),
                )



import contextlib
import os
import tempfile
from textwrap import dedent

import pytest
import yaml

from tavern._core.exceptions import BadSchemaError
from tavern._core.loader import load_single_document_yaml
from tavern._core.schema.files import verify_tests


@pytest.fixture(name="test_dict")
def fix_test_dict():
    text = dedent(
        """
    ---
    test_name: Make sure server doubles number properly

    stages:
      - name: Make sure number is returned correctly
        request:
          url: http://localhost:5000/double
          json:
            number: 5
          method: POST
          headers:
            content-type: application/json
        response:
          status_code: 200
          json:
            double: 10
    """
    )

    as_dict = yaml.load(text, Loader=yaml.SafeLoader)

    return as_dict


class TestJSON:
    def test_simple_json_body(self, test_dict):
        """Simple json dict in request and response"""
        verify_tests(test_dict)

    def test_json_list_request(self, test_dict):
        """Request contains a list"""
        test_dict["stages"][0]["request"]["json"] = [1, "text", -1]

        verify_tests(test_dict)

    def test_json_list_response(self, test_dict):
        """Response contains a list"""
        test_dict["stages"][0]["response"]["json"] = [1, "text", -1]

        verify_tests(test_dict)


class TestHeaders:
    def test_header_request_list(self, test_dict):
        """Headers must always be a dict"""
        test_dict["stages"][0]["request"]["headers"] = [1, "text", -1]

        with pytest.raises(BadSchemaError):
            verify_tests(test_dict)

    def test_headers_response_list(self, test_dict):
        """Headers must always be a dict"""
        test_dict["stages"][0]["response"]["headers"] = [1, "text", -1]

        with pytest.raises(BadSchemaError):
            verify_tests(test_dict)


class TestParameters:
    def test_header_request_list(self, test_dict):
        """Parameters must always be a dict"""
        test_dict["stages"][0]["request"]["params"] = [1, "text", -1]

        with pytest.raises(BadSchemaError):
            verify_tests(test_dict)


class TestTimeout:
    @pytest.mark.parametrize("incorrect_value", ("abc", True, {"a": 2}, [1, 2, 3]))
    def test_timeout_single_fail(self, test_dict, incorrect_value):
        """Timeout must be a list of floats or a float"""
        test_dict["stages"][0]["request"]["timeout"] = incorrect_value

        with pytest.raises(BadSchemaError):
            verify_tests(test_dict)

    @pytest.mark.parametrize("incorrect_value", ("abc", True, None, {"a": 2}))
    def test_timeout_tuple_fail(self, test_dict, incorrect_value):
        """Timeout must be a list of floats or a float"""
        test_dict["stages"][0]["request"]["timeout"] = [1, incorrect_value]

        with pytest.raises(BadSchemaError):
            verify_tests(test_dict)

        test_dict["stages"][0]["request"]["timeout"] = [incorrect_value, 1]

        with pytest.raises(BadSchemaError):
            verify_tests(test_dict)


class TestCert:
    @pytest.mark.parametrize("correct_value", ("a", ["a", "b"]))
    def test_cert_as_string_tuple_list(self, test_dict, correct_value):
        test_dict["stages"][0]["request"]["cert"] = correct_value
        verify_tests(test_dict)

    @pytest.mark.parametrize(
        "incorrect_value", (None, True, {}, ("a", "b", "c"), [], ["a"], ["a", "b", "c"])
    )
    def test_cert_as_tuple(self, test_dict, incorrect_value):
        test_dict["stages"][0]["request"]["cert"] = incorrect_value
        with pytest.raises(BadSchemaError):
            verify_tests(test_dict)


class TestVerify:
    @pytest.mark.parametrize("correct_value", ("a", True, False))
    def test_verify_with_string_boolean(self, test_dict, correct_value):
        test_dict["stages"][0]["request"]["verify"] = correct_value
        verify_tests(test_dict)

    @pytest.mark.parametrize("incorrect_value", (None, 1, {}, [], ("a", "b")))
    def test_verify_with_incorrect_value(self, test_dict, incorrect_value):
        test_dict["stages"][0]["request"]["verify"] = incorrect_value
        with pytest.raises(BadSchemaError):
            verify_tests(test_dict)


class TestBadSchemaAtCollect:
    """Some errors happen at collection time - harder to test"""

    @staticmethod
    @contextlib.contextmanager
    def wrapfile_nondict(to_wrap):
        with tempfile.NamedTemporaryFile(
            suffix=".yaml", prefix="test_", delete=False
        ) as wrapped_tmp:
            # put into a file
            wrapped_tmp.write(to_wrap.encode("utf8"))
            wrapped_tmp.close()

            try:
                yield wrapped_tmp.name
            finally:
                os.remove(wrapped_tmp.name)

    def test_empty_dict_val(self):
        """Defining an empty mapping value is not allowed"""

        text = dedent(
            """
        ---

        test_name: Test cannot send a set

        stages:
          - name: match top level
            request:
              url: "{host}/fake_dictionary"
              method: GET
              json: {a, b}
            response:
              status_code: 200
              json:
                top: !anything
        """
        )

        with TestBadSchemaAtCollect.wrapfile_nondict(text) as filename:
            with pytest.raises(BadSchemaError):
                load_single_document_yaml(filename)

    def test_empty_list_val(self):
        """Defining an empty list value is not allowed"""

        text = dedent(
            """
        ---

        test_name: Test cannot send a set

        stages:
          - name: match top level
            request:
              url: "{host}/fake_dictionary"
              method: GET
              json:
                - a
                -
                - b
            response:
              status_code: 200
              json:
                top: !anything
        """
        )

        with TestBadSchemaAtCollect.wrapfile_nondict(text) as filename:
            with pytest.raises(BadSchemaError):
                load_single_document_yaml(filename)



import copy
import dataclasses
import functools
import logging
import pathlib
from contextlib import ExitStack
from copy import deepcopy
from typing import Dict, List, Mapping, MutableMapping

import box

from tavern._core import exceptions
from tavern._core.plugins import (
    PluginHelperBase,
    get_expected,
    get_extra_sessions,
    get_request_type,
    get_verifiers,
)
from tavern._core.strict_util import StrictLevel

from .dict_util import format_keys, get_tavern_box
from .pytest import call_hook
from .pytest.config import TestConfig
from .report import attach_stage_content, wrap_step
from .strtobool import strtobool
from .testhelpers import delay, retry
from .tincture import Tinctures, get_stage_tinctures

logger: logging.Logger = logging.getLogger(__name__)


def _resolve_test_stages(test_spec: Mapping, available_stages: Mapping):
    # Need to get a final list of stages in the tests (resolving refs)
    test_stages = []
    for raw_stage in test_spec["stages"]:
        stage = raw_stage
        if stage.get("type") == "ref":
            if "id" in stage:
                ref_id = stage["id"]
                if ref_id in available_stages:
                    # Make sure nothing downstream can change the globally
                    # defined stage. Just give the test a local copy.
                    stage = deepcopy(available_stages[ref_id])
                    logger.debug("found stage reference: %s", ref_id)
                else:
                    logger.error("Bad stage: unknown stage referenced: %s", ref_id)
                    raise exceptions.InvalidStageReferenceError(
                        f"Unknown stage reference: {ref_id}"
                    )
            else:
                logger.error("Bad stage: 'ref' type must specify 'id'")
                raise exceptions.BadSchemaError("'ref' stage type must specify 'id'")
        test_stages.append(stage)

    return test_stages


def _get_included_stages(
    tavern_box: box.Box,
    test_block_config: TestConfig,
    test_spec: Mapping,
    available_stages: List[dict],
) -> List[dict]:
    """
    Get any stages which were included via config files which will be available
    for use in this test

    Args:
        tavern_box: Available parameters for fomatting at this point
        test_block_config: Current test config dictionary
        test_spec: Specification for current test
        available_stages: List of stages which already exist

    Returns:
        Fully resolved stages
    """

    def stage_ids(s):
        return [i["id"] for i in s]

    if test_spec.get("includes"):
        # Need to do this separately here so there is no confusion between global and included stages
        for included in test_spec["includes"]:
            for stage in included.get("stages", {}):
                if stage["id"] in stage_ids(available_stages):
                    raise exceptions.DuplicateStageDefinitionError(
                        "Stage id '{}' defined in stage-included test which was already defined in global configuration".format(
                            stage["id"]
                        )
                    )

        included_stages = []  # type: ignore

        for included in test_spec["includes"]:
            if "variables" in included:
                formatted_include = format_keys(included["variables"], tavern_box)
                test_block_config.variables.update(formatted_include)

            for stage in included.get("stages", []):
                if stage["id"] in stage_ids(included_stages):
                    raise exceptions.DuplicateStageDefinitionError(
                        "Stage with specified id already defined: {}".format(
                            stage["id"]
                        )
                    )
                included_stages.append(stage)
    else:
        included_stages = []

    return included_stages


def run_test(
    in_file: pathlib.Path,
    test_spec: MutableMapping,
    global_cfg: TestConfig,
) -> None:
    """Run a single tavern test

    Note that each tavern test can consist of multiple requests (log in,
     create, update, delete, etc).

    The global configuration is copied and used as an initial configuration for
    this test. Any values which are saved from any tests are saved into this
    test block and can be used for formatting in later stages in the test.

    Args:
        in_file: filename containing this test
        test_spec: The specification for this test
        global_cfg: Any global configuration for this test

    Raises:
        TavernException: If any of the tests failed
    """

    # Initialise test config for this test with the global configuration before
    # starting
    test_block_config = global_cfg.copy()
    default_global_strictness = global_cfg.strict

    tavern_box = get_tavern_box()

    if not test_spec:
        logger.warning("Empty test block in %s", in_file)
        return

    # Get included stages and resolve any into the test spec dictionary
    available_stages = test_block_config.stages
    included_stages = _get_included_stages(
        tavern_box, test_block_config, test_spec, available_stages
    )
    all_stages = {s["id"]: s for s in available_stages + included_stages}
    test_spec["stages"] = _resolve_test_stages(test_spec, all_stages)

    test_block_config.variables["tavern"] = tavern_box["tavern"]

    test_block_name = test_spec["test_name"]

    logger.info("Running test : %s", test_block_name)

    with ExitStack() as stack:
        sessions = get_extra_sessions(test_spec, test_block_config)

        for name, session in sessions.items():
            logger.debug("Entering context for %s", name)
            stack.enter_context(session)

        def getonly(stage):
            o = stage.get("only")
            if o is None:
                return False
            elif isinstance(o, bool):
                return o
            else:
                return strtobool(o)

        has_only = any(getonly(stage) for stage in test_spec["stages"])

        runner = _TestRunner(
            default_global_strictness, sessions, test_block_config, test_spec
        )

        try:
            # Run tests in a path in order
            for idx, stage in enumerate(test_spec["stages"]):
                if stage.get("skip"):
                    continue
                if has_only and not getonly(stage):
                    continue

                runner.run_stage(idx, stage)

                if getonly(stage):
                    break
        finally:
            finally_stages = test_spec.get("finally", [])
            if not isinstance(finally_stages, list):
                raise exceptions.BadSchemaError(
                    f"finally block should be a list of dicts but was {type(finally_stages)}"
                )
            for idx, stage in enumerate(finally_stages):
                if not isinstance(stage, dict):
                    raise exceptions.BadSchemaError(
                        f"finally block should be a dict but was {type(stage)}"
                    )
                runner.run_stage(idx, stage, is_final=True)


def _calculate_stage_strictness(
    stage: dict, test_block_config: TestConfig, test_spec: Mapping
) -> StrictLevel:
    """Figure out the strictness for this stage

    Can be overridden per stage, or per test

    Priority is global (see pytest _core.util file) <= test <= stage
    """
    stage_options = None
    new_strict = test_block_config.strict

    if test_spec.get("strict", None) is not None:
        stage_options = test_spec["strict"]
        logger.debug("Getting test level strict setting: %s", stage_options)

    stage_strictness_set = None

    def update_stage_options(new_option):
        if stage_strictness_set:
            raise exceptions.DuplicateStrictError
        logger.debug("Setting stage level strict setting: %s", new_option)
        return new_option

    if stage.get("response", {}).get("strict", None) is not None:
        stage_strictness_set = stage_options = update_stage_options(
            stage["response"]["strict"]
        )

    mqtt_response = stage.get("mqtt_response", None)
    if mqtt_response is not None:
        if isinstance(mqtt_response, dict):
            if mqtt_response.get("strict", None) is not None:
                stage_strictness_set = stage_options = update_stage_options(
                    stage["mqtt_response"]["strict"]
                )
        elif isinstance(mqtt_response, list):
            for response in mqtt_response:
                if response.get("strict", None) is not None:
                    stage_strictness_set = stage_options = update_stage_options(
                        response["strict"]
                    )
        else:
            raise exceptions.BadSchemaError(
                f"mqtt_response was invalid type {type(mqtt_response)}"
            )

    if stage_options is not None:
        if stage_options is True:
            new_strict = StrictLevel.all_on()
        elif stage_options is False:
            new_strict = StrictLevel.all_off()
        else:
            new_strict = StrictLevel.from_options(stage_options)
    else:
        logger.debug("Global default strictness used for this stage")

    logger.debug("Strict key checking for this stage is '%s'", test_block_config.strict)

    return new_strict


@dataclasses.dataclass(frozen=True)
class _TestRunner:
    default_global_strictness: StrictLevel
    sessions: Dict[str, PluginHelperBase]
    test_block_config: TestConfig
    test_spec: Mapping

    def run_stage(self, idx: int, stage, *, is_final: bool = False) -> None:
        tinctures = get_stage_tinctures(stage, self.test_spec)

        stage_config = self.test_block_config.with_strictness(
            self.default_global_strictness
        )
        stage_config = stage_config.with_strictness(
            _calculate_stage_strictness(stage, stage_config, self.test_spec)
        )
        # Wrap run_stage with retry helper
        run_stage_with_retries = retry(stage, stage_config)(self.wrapped_run_stage)
        partial = functools.partial(
            run_stage_with_retries, stage, stage_config, tinctures
        )
        allure_name = "Stage {}: {}".format(
            idx, format_keys(stage["name"], stage_config.variables)
        )
        step = wrap_step(allure_name, partial)

        try:
            step()
        except exceptions.TavernException as e:
            e.stage = stage
            e.test_block_config = stage_config
            e.is_final = is_final
            raise

    def wrapped_run_stage(
        self, stage: dict, stage_config: TestConfig, tinctures: Tinctures
    ) -> None:
        """Run one stage from the test

        Args:
            stage: specification of stage to be run
            stage_config: available variables for test
            tinctures: tinctures for this stage/test
        """
        stage = copy.deepcopy(stage)
        name = stage["name"]

        attach_stage_content(stage)

        r = get_request_type(stage, stage_config, self.sessions)

        tavern_box = stage_config.variables["tavern"]
        tavern_box.update(request_vars=r.request_vars)

        expected = get_expected(stage, stage_config, self.sessions)

        delay(stage, "before", stage_config.variables)

        logger.info("Running stage : %s", name)

        call_hook(
            stage_config,
            "pytest_tavern_beta_before_every_request",
            request_args=r.request_vars,
        )

        verifiers = get_verifiers(stage, stage_config, self.sessions, expected)

        tinctures.start_tinctures(stage)

        response = r.run()

        tinctures.end_tinctures(expected, response)

        for response_type, response_verifiers in verifiers.items():
            logger.debug("Running verifiers for %s", response_type)
            for v in response_verifiers:
                saved = v.verify(response)
                stage_config.variables.update(saved)

        tavern_box.pop("request_vars")
        delay(stage, "after", stage_config.variables)



import contextlib
import functools
import logging
import os
import re
import string
import typing
from collections.abc import Collection
from typing import Any, Dict, Iterator, List, Mapping, Optional, Tuple, Union

import box
import jmespath
from box.box import Box

from tavern._core import exceptions
from tavern._core.loader import (
    ANYTHING,
    ForceIncludeToken,
    RegexSentinel,
    TypeConvertToken,
    TypeSentinel,
)

from .formatted_str import FormattedString
from .strict_util import StrictSetting, StrictSettingKinds, extract_strict_setting

logger: logging.Logger = logging.getLogger(__name__)


def _check_and_format_values(to_format: str, box_vars: Mapping[str, Any]) -> str:
    formatter = string.Formatter()
    would_format = formatter.parse(to_format)

    for _, field_name, _, _ in would_format:
        if field_name is None:
            continue

        try:
            would_replace = formatter.get_field(field_name, [], box_vars)[0]
        except KeyError as e:
            logger.error(
                "Failed to resolve string '%s' with variables '%s'", to_format, box_vars
            )
            logger.error("Key(s) not found in format: %s", field_name)
            raise exceptions.MissingFormatError(field_name) from e
        except IndexError as e:
            logger.error("Empty format values are invalid")
            raise exceptions.MissingFormatError(field_name) from e
        else:
            if not isinstance(would_replace, (str, int, float)):
                logger.warning(
                    "Formatting '%s' will result in it being coerced to a string (it is a %s)",
                    field_name,
                    type(would_replace),
                )

    return to_format.format(**box_vars)


def _attempt_find_include(to_format: str, box_vars: box.Box) -> Optional[str]:
    formatter = string.Formatter()
    would_format = list(formatter.parse(to_format))

    yaml_tag = ForceIncludeToken.yaml_tag

    if len(would_format) != 1:
        raise exceptions.InvalidFormattedJsonError(
            "When using {}, there can only be one exactly format value, but got {}".format(
                yaml_tag, len(would_format)
            )
        )

    (_, field_name, format_spec, conversion) = would_format[0]

    if field_name is None:
        raise exceptions.InvalidFormattedJsonError(
            f"Invalid string used for {yaml_tag}"
        )

    pattern = r"{" + field_name + r".*}"

    if not re.match(pattern, to_format):
        raise exceptions.InvalidFormattedJsonError(
            f"Invalid format specifier '{to_format}' for {yaml_tag}"
        )

    if format_spec:
        logger.warning(
            "Conversion specifier '%s' will be ignored for %s", format_spec, to_format
        )

    would_replace = formatter.get_field(field_name, [], box_vars)[0]

    if conversion is None:
        return would_replace

    return formatter.convert_field(would_replace, conversion)


T = typing.TypeVar("T", str, Dict, List, Tuple)


def format_keys(
    val: T,
    variables: Union[Mapping, Box],
    *,
    no_double_format: bool = True,
    dangerously_ignore_string_format_errors: bool = False,
) -> T:
    """recursively format a dictionary with the given values

    Args:
        val: Input thing to format
        variables: Dictionary of keys to format it with
        no_double_format: Whether to use the 'inner formatted string' class to avoid double formatting
            This is required if passing something via pytest-xdist, such as markers:
            https://github.com/taverntesting/tavern/issues/431
        dangerously_ignore_string_format_errors: whether to ignore any string formatting errors. This will result
            in broken output, only use for debugging purposes.

    Raises:
        MissingFormatError: if a format variable was not found in variables

    Returns:
        recursively formatted values
    """
    format_keys_ = functools.partial(
        format_keys,
        dangerously_ignore_string_format_errors=dangerously_ignore_string_format_errors,
    )

    if not isinstance(variables, Box):
        box_vars = Box(variables)
    else:
        box_vars = variables

    if isinstance(val, dict):
        return {key: format_keys_(val[key], box_vars) for key in val}
    elif isinstance(val, tuple):
        return tuple(format_keys_(item, box_vars) for item in val)
    elif isinstance(val, list):
        return [format_keys_(item, box_vars) for item in val]
    elif isinstance(val, FormattedString):
        logger.debug("Already formatted %s, not double-formatting", val)
    elif isinstance(val, str):
        formatted = val
        try:
            formatted = _check_and_format_values(val, box_vars)
        except exceptions.MissingFormatError:
            if not dangerously_ignore_string_format_errors:
                raise

        if no_double_format:
            formatted = FormattedString(formatted)  # type: ignore

        return formatted
    elif isinstance(val, TypeConvertToken):
        logger.debug("Got type convert token '%s'", val)
        if isinstance(val, ForceIncludeToken):
            return _attempt_find_include(val.value, box_vars)
        else:
            value = format_keys_(val.value, box_vars)
            return val.constructor(value)
    else:
        logger.debug("Not formatting something of type '%s'", type(val))

    return val


def recurse_access_key(data: Union[List, Mapping], query: str) -> Any:
    """
    Search for something in the given data using the given query.

    Example:

        >>> recurse_access_key({"a": "b"}, "a")
        'b'
        >>> recurse_access_key({"a": {"b": ["c", "d"]}}, "a.b[0]")
        'c'

    Args:
        data: Data to search in
        query: Query to run

    Raises:
        JMESError: if there was an error parsing the query

    Returns:
        Whatever was found by the search
    """

    try:
        from_jmespath = jmespath.search(query, data)
    except jmespath.exceptions.ParseError as e:
        logger.error("Error parsing JMES query")

        try:
            _deprecated_recurse_access_key(data, query.split("."))
        except (IndexError, KeyError):
            logger.debug("Nothing found searching using old method")
        else:
            # If we found a key using 'old' style searching
            logger.warning(
                "Something was found using 'old style' searching in the response - please change the query to use jmespath instead - see http://jmespath.org/ for more information"
            )

        raise exceptions.JMESError("Invalid JMES query") from e

    return from_jmespath


def _deprecated_recurse_access_key(
    current_val: Union[List, Mapping], keys: List
) -> Any:
    """Given a list of keys and a dictionary, recursively access the dicionary
    using the keys until we find the key its looking for

    If a key is an integer, it will convert it and use it as a list index

    Example:

        >>> _deprecated_recurse_access_key({"a": "b"}, ["a"])
        'b'
        >>> _deprecated_recurse_access_key({"a": {"b": ["c", "d"]}}, ["a", "b", "0"])
        'c'

    Args:
        current_val: current dictionary we have recursed into
        keys: list of str/int of subkeys

    Raises:
        IndexError: list index not found in data
        KeyError: dict key not found in data

    Returns:
        value of subkey in dict
    """
    logger.debug("Recursively searching for '%s' in '%s'", keys, current_val)

    if not keys:
        return current_val
    else:
        current_key = keys.pop(0)

        with contextlib.suppress(ValueError):
            current_key = int(current_key)

        try:
            return _deprecated_recurse_access_key(current_val[current_key], keys)
        except (IndexError, KeyError, TypeError) as e:
            logger.error(
                "%s accessing data - looking for '%s' in '%s'",
                type(e).__name__,
                current_key,
                current_val,
            )
            raise


def deep_dict_merge(initial_dct: Dict, merge_dct: Mapping) -> dict:
    """Recursive dict merge. Instead of updating only top-level keys,
    dict_merge recurses down into dicts nested to an arbitrary depth
    and returns the merged dict. Keys values present in merge_dct take
    precedence over values in initial_dct.
    Modified from: https://gist.github.com/angstwad/bf22d1822c38a92ec0a9

    Params:
        initial_dct: dict onto which the merge is executed
        merge_dct: dct merged into dct

    Returns:
        recursively merged dict
    """
    dct = initial_dct.copy()

    for k in merge_dct:
        if k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], Mapping):
            dct[k] = deep_dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]

    return dct


def check_expected_keys(expected: Collection, actual: Collection) -> None:
    """Check that a set of expected keys is a superset of the actual keys

    Args:
        expected: keys we expect
        actual: keys we have got from the input

    Raises:
        UnexpectedKeysError: If not actual <= expected
    """
    expected = set(expected)
    keyset = set(actual)

    if not keyset <= expected:
        unexpected = keyset - expected

        logger.debug("Valid keys = %s, actual keys = %s", expected, keyset)

        msg = f"Unexpected keys {unexpected}"
        logger.error(msg)
        raise exceptions.UnexpectedKeysError(msg)


def yield_keyvals(block: Union[List, Dict]) -> Iterator[Tuple[List, str, str]]:
    """Return indexes, keys and expected values for matching recursive keys

    Given a list or dict, return a 3-tuple of the 'split' key (key split on
    dots), the original key, and the expected value. If the input is a list, it
    is enumerated so the 'keys' are just [0, 1, 2, ...]

    Example:

        Matching a dictionary with a couple of keys:

        >>> gen = yield_keyvals({"a": {"b": "c"}})
        >>> next(gen)
        (['a'], 'a', {'b': 'c'})

        Matching nested key access:

        >>> gen = yield_keyvals({"a.b.c": "d"})
        >>> next(gen)
        (['a', 'b', 'c'], 'a.b.c', 'd')

        Matching a list of items:

        >>> gen = yield_keyvals(["a", "b", "c"])
        >>> next(gen)
        (['0'], '0', 'a')
        >>> next(gen)
        (['1'], '1', 'b')
        >>> next(gen)
        (['2'], '2', 'c')

    Args:
        block: input matches

    Yields:
        iterable of (key split on dots, key, expected value)
    """
    if isinstance(block, dict):
        for joined_key, expected_val in block.items():
            split_key = joined_key.split(".")
            yield split_key, joined_key, expected_val
    else:
        for idx, val in enumerate(block):
            sidx = str(idx)
            yield [sidx], sidx, val


Checked = typing.TypeVar("Checked", Dict, Collection, str)


def check_keys_match_recursive(
    expected_val: Checked,
    actual_val: Checked,
    keys: List[Union[str, int]],
    strict: StrictSettingKinds = True,
) -> None:
    """Utility to recursively check response values

    expected and actual both have to be of the same type or it will raise an
    error.

    Example:

        >>> check_keys_match_recursive({"a": {"b": "c"}}, {"a": {"b": "c"}}, []) is None
        True
        >>> check_keys_match_recursive(
        ...     {"a": {"b": "c"}}, {"a": {"b": "d"}}, []
        ... )  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
          File "/home/michael/code/tavern/tavern/tavern/_core.util/dict_util.py", line 223, in check_keys_match_recursive
        tavern._core.exceptions.KeyMismatchError: Key mismatch: (expected["a"]["b"] = 'c', actual["a"]["b"] = 'd')

    Todo:
        This could be turned into a single-dispatch function for cleaner
        code and to remove a load of the isinstance checks

    Args:
        expected_val: expected value
        actual_val: actual value
        keys: any keys which have been recursively parsed to get to this
            point. Used for debug output.
        strict: Whether 'strict' key checking should be done. If this is
            False, a mismatch in dictionary keys between the expected and the
            actual values will not raise an error (but a mismatch in value will
            raise an error)

    Raises:
        KeyMismatchError: expected_val and actual_val did not match
    """

    def full_err():
        """Get error in the format:

        a["b"]["c"] = 4, b["b"]["c"] = {'key': 'value'}
        """

        def _format_err(which):
            return "{}{}".format(which, "".join(f'["{key}"]' for key in keys))

        e_formatted = _format_err("expected")
        a_formatted = _format_err("actual")
        return "{} = '{}' (type = {}), {} = '{}' (type = {})".format(
            e_formatted,
            expected_val,
            type(expected_val),
            a_formatted,
            actual_val,
            type(actual_val),
        )

    actual_type = type(actual_val)

    if expected_val is ANYTHING:
        # Match anything. We could just early exit here but having the debug
        # logging below is useful
        expected_matches = True
    elif isinstance(expected_val, TypeSentinel):
        # If the 'expected' type is actually just a sentinel for another type,
        # then it should match
        expected_matches = expected_val.constructor == actual_type
    else:
        # Normal matching
        expected_matches = (
            # If they are the same type
            isinstance(expected_val, actual_type)
            or
            # Handles the case where, for example, the 'actual type' returned by
            # a custom backend returns an OrderedDict, which is a subclass of
            # dict but will raise a confusing error if the contents are
            # different
            issubclass(actual_type, type(expected_val))
        )

    strict_bool, strict_setting = extract_strict_setting(strict)

    try:
        assert actual_val == expected_val  # noqa
    except AssertionError as e:
        # At this point, there is likely to be an error unless we're using any
        # of the type sentinels

        if expected_val is not ANYTHING:
            if not expected_matches:
                if isinstance(expected_val, RegexSentinel):
                    msg = "Expected a string to match regex '{}' ({})".format(
                        expected_val.compiled, full_err()
                    )
                else:
                    msg = (
                        "Type of returned data was different than expected ({})".format(
                            full_err()
                        )
                    )

                raise exceptions.KeyMismatchError(msg) from e

        if isinstance(expected_val, dict):
            ekeys = set(expected_val.keys())
            akeys = set(actual_val.keys())  # type:ignore

            if akeys != ekeys:
                extra_actual_keys = akeys - ekeys
                extra_expected_keys = ekeys - akeys

                msg = ""
                if extra_actual_keys:
                    msg += f" - Extra keys in response: {extra_actual_keys}"
                if extra_expected_keys:
                    msg += " - Keys missing from response: {}".format(
                        extra_expected_keys
                    )

                full_msg = "Structure of returned data was different than expected {} ({})".format(
                    msg, full_err()
                )

                # If there are more keys in 'expected' compared to 'actual',
                # this is still a hard error and we shouldn't continue
                if extra_expected_keys or strict_bool:
                    raise exceptions.KeyMismatchError(full_msg) from e
                else:
                    logger.debug(
                        "Mismatch in returned data, continuing due to strict=%s: %s",
                        strict_bool,
                        full_msg,
                        exc_info=True,
                    )

            # If strict is True, an error will be raised above. If not, recurse
            # through both sets of keys and just ignore missing ones
            to_recurse = akeys | ekeys

            for key in to_recurse:
                try:
                    check_keys_match_recursive(
                        expected_val[key],
                        actual_val[key],  # type:ignore
                        keys + [key],
                        strict,
                    )
                except KeyError:
                    logger.debug(
                        "Skipping comparing missing key '%s' due to strict=%s",
                        key,
                        strict_bool,
                    )
        elif isinstance(expected_val, list):
            if not strict_bool:
                missing = []

                actual_iter = iter(actual_val)

                # Iterate over list items to see if any of them match _IN ORDER_
                for i, e_val in enumerate(expected_val):
                    while 1:
                        try:
                            current_response_val = next(actual_iter)
                        except StopIteration:
                            # Still iterating checking for a value, but ran out of response values
                            logger.debug("Ran out of list response items to check")
                            missing.append(e_val)
                            break
                        else:
                            logger.debug(
                                "Got '%s' from response to check against '%s' from expected",
                                current_response_val,
                                e_val,
                            )

                        # Found one - check if it matches
                        try:
                            check_keys_match_recursive(
                                e_val, current_response_val, keys + [i], strict
                            )
                        except exceptions.KeyMismatchError:
                            # Doesn't match what we're looking for
                            logger.debug(
                                "%s did not match next response value %s",
                                e_val,
                                current_response_val,
                            )
                        else:
                            logger.debug("'%s' present in response", e_val)
                            if strict_setting == StrictSetting.LIST_ANY_ORDER:
                                actual_iter = iter(actual_val)
                            break

                if missing:
                    msg = f"List item(s) not present in response: {missing}"
                    raise exceptions.KeyMismatchError(msg) from e

                logger.debug("All expected list items present")
            else:
                if len(expected_val) != len(actual_val):
                    raise exceptions.KeyMismatchError(
                        "Length of returned list was different than expected - expected {} items from got {} ({}".format(
                            len(expected_val), len(actual_val), full_err()
                        )
                    ) from e

                for i, (e_val, a_val) in enumerate(zip(expected_val, actual_val)):
                    try:
                        check_keys_match_recursive(e_val, a_val, keys + [i], strict)
                    except exceptions.KeyMismatchError as sub_e:
                        # This should _ALWAYS_ raise an error (unless the reason it didn't match was the
                        # 'anything' sentinel), but it will be more obvious where the error came from
                        # (in python 3 at least) and will take ANYTHING into account
                        raise sub_e from e
        elif expected_val is ANYTHING:
            logger.debug("Actual value = '%s' - matches !anything", actual_val)
        elif isinstance(expected_val, TypeSentinel) and expected_matches:
            if isinstance(expected_val, RegexSentinel):
                if not expected_val.passes(actual_val):
                    raise exceptions.KeyMismatchError(
                        f"Regex mismatch: ({full_err()})"
                    ) from e

            logger.debug(
                "Actual value = '%s' - matches !any%s",
                actual_val,
                expected_val.constructor,
            )
        else:
            raise exceptions.KeyMismatchError(f"Key mismatch: ({full_err()})") from e


def get_tavern_box() -> box.Box:
    """Get the 'tavern' box"""
    return Box({"tavern": {"env_vars": dict(os.environ)}})



import time
from typing import Dict
from unittest.mock import MagicMock, Mock, patch

import paho.mqtt.client as paho
import pytest

from tavern._core import exceptions
from tavern._plugins.mqtt.client import MQTTClient, _handle_tls_args, _Subscription
from tavern._plugins.mqtt.request import MQTTRequest


def test_host_required():
    """Always needs a host, but it's the only required key"""
    with pytest.raises(exceptions.MissingKeysError):
        MQTTClient()

    args = {"connect": {"host": "localhost"}}

    MQTTClient(**args)


@pytest.fixture(name="fake_client")
def fix_fake_client():
    args = {"connect": {"host": "localhost", "timeout": 0.6}}

    mqtt_client = MQTTClient(**args)

    mqtt_client._subscribed[2] = _Subscription("abc")
    mqtt_client._subscription_mappings["abc"] = 2

    return mqtt_client


class TestClient:
    def test_no_queue(self, fake_client):
        """Trying to fetch from a nonexistent queue raised exception"""

        with pytest.raises(exceptions.MQTTTopicException):
            fake_client.message_received("", 0)

    def test_no_message(self, fake_client):
        """No message in queue returns None"""

        assert fake_client.message_received("abc", 0) is None

    def test_message_queued(self, fake_client):
        """Returns message in queue"""

        message = "abc123"

        fake_client._userdata["_subscribed"][2].queue.put(message)
        assert fake_client.message_received("abc", 0) == message

    def test_context_connection_failure(self, fake_client):
        """Unable to connect on __enter__ raises MQTTError"""

        fake_client._connect_timeout = 0.3

        with patch.object(fake_client._client, "loop_start"):
            with pytest.raises(exceptions.MQTTError):
                with fake_client:
                    pass

    def test_context_connection_success(self, fake_client):
        """returns self on success"""

        with patch.object(fake_client._client, "loop_start"), patch.object(
            fake_client._client, "connect_async"
        ):
            fake_client._client._state = paho.mqtt_cs_connected
            with fake_client as x:
                assert fake_client == x

    def test_assert_message_published_error(self, fake_client):
        """Error waiting for it to publish"""

        class FakeMessage(paho.MQTTMessageInfo):
            def wait_for_publish(self, timeout=None):
                raise RuntimeError

            rc = 1

        with patch.object(fake_client._client, "subscribe"), patch.object(
            fake_client._client, "publish", return_value=FakeMessage(10)
        ):
            with pytest.raises(exceptions.MQTTError):
                fake_client.publish("abc", "123")

    def test_assert_message_published_failure(self, fake_client: MQTTClient):
        """If it couldn't publish the message, error out"""

        class FakeMessage(paho.MQTTMessageInfo):
            def wait_for_publish(self, timeout=None):
                return

            def is_published(self):
                return False

            rc = 1

        with patch.object(fake_client._client, "subscribe"), patch.object(
            fake_client._client, "publish", return_value=FakeMessage(10)
        ):
            with pytest.raises(exceptions.MQTTError):
                fake_client.publish("abc", "123")

    def test_assert_message_published_delay(self, fake_client):
        """Published but only after a small delay"""

        class FakeMessage(paho.MQTTMessageInfo):
            def wait_for_publish(self, timeout=None):
                time.sleep(0.5)

            def is_published(self):
                return True

            rc = 1

        with patch.object(fake_client._client, "subscribe"), patch.object(
            fake_client._client, "publish", return_value=FakeMessage(10)
        ):
            fake_client.publish("abc", "123")

    def test_assert_message_published_unknown_err(self, fake_client):
        """Same, but with an unknown error code"""

        class FakeMessage(paho.MQTTMessageInfo):
            def is_published(self):
                return False

            rc = 2342423

        with patch.object(fake_client._client, "subscribe"), patch.object(
            fake_client._client, "publish", return_value=FakeMessage(10)
        ):
            with pytest.raises(exceptions.MQTTError):
                fake_client.publish("abc", "123")


class TestTLS:
    def test_missing_cert_gives_error(self):
        """Missing TLS cert gives an error"""
        args = {"certfile": "/lcliueurhug/ropko3kork32"}

        with pytest.raises(exceptions.MQTTTLSError):
            _handle_tls_args(args)

    def test_disabled_tls(self):
        """Even if there are other invalid options, disable tls and early exit
        without checking other args
        """
        args = {"certfile": "/lcliueurhug/ropko3kork32", "enable": False}

        parsed_args = _handle_tls_args(args)
        assert not parsed_args

    def test_invalid_tls_ver(self):
        """Bad tls versions raise exception"""
        args = {"tls_version": "custom_tls"}

        with pytest.raises(exceptions.MQTTTLSError):
            _handle_tls_args(args)


@pytest.fixture(name="req")
def fix_example_request():
    spec = {"topic": "{request_topic:s}", "payload": "abc123"}

    return spec.copy()


class TestRequests:
    def test_unknown_fields(self, req, includes):
        """Unkown args should raise an error"""
        req["fodokfowe"] = "Hello"

        with pytest.raises(exceptions.UnexpectedKeysError):
            MQTTRequest(Mock(), req, includes)

    def test_missing_format(self, req, includes):
        """All format variables should be present"""
        del includes.variables["request_topic"]

        with pytest.raises(exceptions.MissingFormatError):
            MQTTRequest(Mock(), req, includes)

    def test_correct_format(self, req, includes):
        """All format variables should be present"""
        MQTTRequest(Mock(), req, includes)


class TestSubscription:
    @staticmethod
    def get_mock_client_with(subcribe_action):
        mock_paho = Mock(spec=paho.Client, subscribe=subcribe_action)
        mock_client = Mock(
            spec=MQTTClient,
            _client=mock_paho,
            _subscribed={},
            _subscription_mappings={},
            _subscribe_lock=MagicMock(),
        )
        return mock_client

    def test_handles_subscriptions(self):
        def subscribe_success(topic, *args, **kwargs):
            return (0, 123)

        mock_client = TestSubscription.get_mock_client_with(subscribe_success)

        MQTTClient.subscribe(mock_client, "abc")

        assert mock_client._subscribed[123].topic == "abc"
        assert mock_client._subscribed[123].subscribed is False

    def test_no_subscribe_on_err(self):
        def subscribe_err(topic, *args, **kwargs):
            return (1, 123)

        mock_client = TestSubscription.get_mock_client_with(subscribe_err)

        with pytest.raises(exceptions.MQTTError):
            MQTTClient.subscribe(mock_client, "abc")

        assert mock_client._subscribed == {}

    def test_no_subscribe_on_unrecognised_suback(self):
        def subscribe_success(topic, *args, **kwargs):
            return (0, 123)

        mock_client = TestSubscription.get_mock_client_with(subscribe_success)

        MQTTClient._on_subscribe(mock_client, "abc", {}, 123, 0)

        assert mock_client._subscribed == {}


class TestExtFunctions:
    @pytest.fixture()
    def basic_mqtt_request_args(self) -> Dict:
        return {
            "topic": "/a/b/c",
        }

    def test_basic(self, fake_client, basic_mqtt_request_args, includes):
        MQTTRequest(fake_client, basic_mqtt_request_args, includes)

    def test_ext_function_bad(self, fake_client, basic_mqtt_request_args, includes):
        basic_mqtt_request_args["json"] = {"$ext": "kk"}

        with pytest.raises(exceptions.InvalidExtFunctionError):
            MQTTRequest(fake_client, basic_mqtt_request_args, includes)

    def test_ext_function_good(self, fake_client, basic_mqtt_request_args, includes):
        basic_mqtt_request_args["json"] = {
            "$ext": {
                "function": "operator:add",
                "extra_args": (1, 2),
            }
        }

        m = MQTTRequest(fake_client, basic_mqtt_request_args, includes)

        assert "payload" in m._publish_args
        assert m._publish_args["payload"] == "3"



import contextlib
import json
import logging
import warnings
from contextlib import ExitStack
from itertools import filterfalse, tee
from typing import Callable, ClassVar, Dict, List, Mapping, Optional
from urllib.parse import quote_plus

import requests
from box.box import Box
from requests.cookies import cookiejar_from_dict
from requests.utils import dict_from_cookiejar

from tavern._core import exceptions
from tavern._core.dict_util import check_expected_keys, deep_dict_merge, format_keys
from tavern._core.extfunctions import update_from_ext
from tavern._core.general import valid_http_methods
from tavern._core.pytest.config import TestConfig
from tavern._core.report import attach_yaml
from tavern._plugins.rest.files import get_file_arguments, guess_filespec
from tavern.request import BaseRequest

logger: logging.Logger = logging.getLogger(__name__)


def get_request_args(rspec: Dict, test_block_config: TestConfig) -> Dict:
    """Format the test spec given values inthe global config

    Todo:
        Add similar functionality to validate/save $ext functions so input
        can be generated from a function

    Args:
        rspec: Test spec
        test_block_config: Test block config

    Returns:
        Formatted test spec

    Raises:
        BadSchemaError: Tried to pass a body in a GET request
    """

    if "method" not in rspec:
        logger.debug("Using default GET method")
        rspec["method"] = "GET"

    if "headers" not in rspec:
        rspec["headers"] = {}

    content_keys = ["data", "json", "files", "file_body"]

    in_request = [c for c in content_keys if c in rspec]
    if len(in_request) > 1:
        # Explicitly raise an error here
        # From requests docs:
        # Note, the json parameter is ignored if either data or files is passed.
        # However, we allow the data + files case, as requests handles it correctly
        if set(in_request) != {"data", "files"}:
            raise exceptions.BadSchemaError(
                "Can only specify one type of request data in HTTP request (tried to "
                "send {})".format(" and ".join(in_request))
            )

    normalised_headers = {k.lower(): v for k, v in rspec["headers"].items()}

    def get_header(name):
        return normalised_headers.get(name, None)

    content_header = get_header("content-type")
    encoding_header = get_header("content-encoding")

    if "files" in rspec:
        if content_header:
            logger.warning(
                "Tried to specify a content-type header while sending multipart files - this will be ignored"
            )
            rspec["headers"] = {
                i: j
                for i, j in normalised_headers.items()
                if i.lower() != "content-type"
            }

    fspec = format_keys(rspec, test_block_config.variables)

    if fspec["method"] not in valid_http_methods:
        raise exceptions.BadSchemaError(
            "Unknown HTTP method {}".format(fspec["method"])
        )

    # If the user is using the file_body key, try to guess what type of file/encoding it is.
    filename = fspec.get("file_body")
    if filename:
        with ExitStack() as stack:
            file_spec, group_name = guess_filespec(filename, stack, test_block_config)

            # Group name doesn't matter here as it's a single file
            if group_name:
                logger.warning(
                    f"'group_name' for the 'file_body' key was specified as '{group_name}' but this will be ignored "
                )

            fspec["file_body"] = filename
            if len(file_spec) == 2:
                logger.debug(
                    "No content type or encoding inferred from file_body for %s",
                    filename,
                )

            if len(file_spec) >= 3:
                inferred_content_type = file_spec[2]
                if content_header:
                    logger.info(
                        "inferred content type '%s' from %s, but using user specified content type '%s'",
                        inferred_content_type,
                        filename,
                        content_header,
                    )
                else:
                    fspec["headers"]["content-type"] = inferred_content_type

            if len(file_spec) == 4:
                inferred_content_encoding = file_spec[3]
                if encoding_header:
                    logger.info(
                        "inferred content encoding '%s' from %s, but using user specified encoding '%s",
                        inferred_content_encoding,
                        filename,
                        encoding_header,
                    )
                else:
                    fspec["headers"].update(**inferred_content_encoding)

    #########################################

    request_args = {}

    def add_request_args(keys, optional):
        for key in keys:
            try:
                request_args[key] = fspec[key]
            except KeyError:
                if optional or (key in request_args):
                    continue

                # This should never happen
                raise

    # Ones that are required and are enforced to be present by the schema
    required_in_file = ["method", "url"]

    add_request_args(["file_body"], True)
    add_request_args(required_in_file, False)
    add_request_args(RestRequest.optional_in_file, True)

    if "auth" in fspec:
        request_args["auth"] = tuple(fspec["auth"])

    if "cert" in fspec:
        if isinstance(fspec["cert"], list):
            request_args["cert"] = tuple(fspec["cert"])

    if "timeout" in fspec:
        # Needs to be a tuple, it being a list doesn't work
        if isinstance(fspec["timeout"], list):
            request_args["timeout"] = tuple(fspec["timeout"])

    # If there's any nested json in parameters, urlencode it
    # if you pass nested json to 'params' then requests silently fails and just
    # passes the 'top level' key, ignoring all the nested json. I don't think
    # there's a standard way to do this, but urlencoding it seems sensible
    # eg https://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter
    # > ...represented in an OAuth 2.0 request as UTF-8 encoded JSON (which ends
    # > up being form-urlencoded when passed as an OAuth parameter)
    for key, value in request_args.get("params", {}).items():
        if not isinstance(value, str):
            if key == "$ext":
                logger.debug("Skipping converting of ext function (%s)", value)
                continue

        if isinstance(value, dict):
            request_args["params"][key] = quote_plus(json.dumps(value))

    optional = {"verify", "stream"}

    for key in optional:
        if key in fspec:
            request_args[key] = fspec[key]

    # TODO
    # requests takes all of these - we need to parse the input to get them
    # "cookies",

    # These verbs _can_ send a body but the body _should_ be ignored according
    # to the specs - some info here:
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
    if request_args["method"] in ["GET", "HEAD", "OPTIONS"]:
        if any(i in request_args for i in ["json", "data"]):
            warnings.warn(  # noqa
                "You are trying to send a body with a HTTP verb that has no semantic use for it",
                RuntimeWarning,
            )

    return request_args


@contextlib.contextmanager
def _set_cookies_for_request(session: requests.Session, request_args: Mapping):
    """
    Possibly reset session cookies for a single request then set them back.
    If no cookies were present in the request arguments, do nothing.

    This does not use try/finally because if it fails then we don't care about
    the cookies anyway

    Args:
        session: Current session
        request_args: current request arguments
    """
    if "cookies" in request_args:
        old_cookies = dict_from_cookiejar(session.cookies)
        session.cookies = cookiejar_from_dict({})
        yield
        session.cookies = cookiejar_from_dict(old_cookies)
    else:
        yield


def _check_allow_redirects(rspec: dict, test_block_config: TestConfig):
    """
    Check for allow_redirects flag in settings/stage

    Args:
        rspec: request dictionary
        test_block_config: config available for test

    Returns:
        Whether to allow redirects for this stage or not
    """
    # By default, don't follow redirects
    allow_redirects = False

    # Then check to see if we should follow redirects based on settings
    global_follow_redirects = test_block_config.follow_redirects
    if global_follow_redirects is not None:
        allow_redirects = global_follow_redirects

    # ... and test flags
    test_follow_redirects = rspec.pop("follow_redirects", None)
    if test_follow_redirects is not None:
        if global_follow_redirects is not None:
            logger.info(
                "Overriding global follow_redirects setting of %s with test-level specification of %s",
                global_follow_redirects,
                test_follow_redirects,
            )
        allow_redirects = test_follow_redirects

    logger.debug("Allow redirects in stage: %s", allow_redirects)

    return allow_redirects


def _read_expected_cookies(
    session: requests.Session, rspec: Mapping, test_block_config: TestConfig
) -> Optional[dict]:
    """
    Read cookies to inject into request, ignoring others which are present

    Args:
        session: session object
        rspec: test spec
        test_block_config: config available for test

    Returns:
        cookies to use in request, if any
    """
    # Need to do this down here - it is separate from getting request args as
    # it depends on the state of the session
    existing_cookies = session.cookies.get_dict()
    cookies_to_use = format_keys(
        rspec.get("cookies", None), test_block_config.variables
    )

    if cookies_to_use is None:
        logger.debug("No cookies specified in request, sending all")
        return None
    elif cookies_to_use in ([], {}):
        logger.debug("Not sending any cookies with request")
        return {}

    def partition(pred, iterable):
        """From itertools documentation"""
        t1, t2 = tee(iterable)
        return list(filterfalse(pred, t1)), list(filter(pred, t2))

    # Cookies are either a single list item, specitying which cookie to send, or
    # a mapping, specifying cookies to override
    expected, extra = partition(lambda x: isinstance(x, dict), cookies_to_use)

    missing = set(expected) - set(existing_cookies.keys())

    if missing:
        logger.error("Missing cookies")
        raise exceptions.MissingCookieError(
            "Tried to use cookies '{}' in request but only had '{}' available".format(
                expected, existing_cookies
            )
        )

    # 'extra' should be a list of dictionaries - merge them into one here
    from_extra = {k: v for mapping in extra for (k, v) in mapping.items()}

    if len(extra) != len(from_extra):
        logger.error("Duplicate cookie override values specified")
        raise exceptions.DuplicateCookieError(
            "Tried to override the value of a cookie multiple times in one request"
        )

    overwritten = [i for i in expected if i in from_extra]

    if overwritten:
        logger.error("Duplicate cookies found in request")
        raise exceptions.DuplicateCookieError(
            "Asked to use cookie {} from previous request but also redefined it as {}".format(
                overwritten, from_extra
            )
        )

    from_cookiejar = {c: existing_cookies.get(c) for c in expected}

    return deep_dict_merge(from_cookiejar, from_extra)


class RestRequest(BaseRequest):
    optional_in_file: ClassVar[List[str]] = [
        "json",
        "data",
        "params",
        "headers",
        "files",
        "timeout",
        "cert",
        # Ideally this would just be passed through but requests seems to error
        # if we pass a list instead of a tuple, so we have to manually convert
        # it further down
        # "auth"
    ]

    def __init__(
        self, session: requests.Session, rspec: dict, test_block_config: TestConfig
    ) -> None:
        """Prepare request

        Args:
            session: existing session
            rspec: test spec
            test_block_config: Any configuration for this the block of
                tests

        Raises:
            UnexpectedKeysError: If some unexpected keys were used in the test
                spec. Only valid keyword args to requests can be passed
        """

        if rspec.pop("clear_session_cookies", False):
            session.cookies.clear_session_cookies()

        expected = {
            "method",
            "url",
            "headers",
            "data",
            "params",
            "auth",
            "json",
            "verify",
            "files",
            "file_body",
            "stream",
            "timeout",
            "cookies",
            "cert",
            # "hooks",
            "follow_redirects",
        }

        check_expected_keys(expected, rspec)

        request_args = get_request_args(rspec, test_block_config)
        update_from_ext(
            request_args,
            RestRequest.optional_in_file + ["url"],
        )

        # Used further down, but pop it asap to avoid unwanted side effects
        file_body: Optional[str] = request_args.pop("file_body", None)

        # If there was a 'cookies' key, set it in the request
        expected_cookies = _read_expected_cookies(session, rspec, test_block_config)
        if expected_cookies is not None:
            logger.debug("Sending cookies %s in request", expected_cookies.keys())
            request_args.update(cookies=expected_cookies)

        # Check for redirects
        request_args.update(
            allow_redirects=_check_allow_redirects(rspec, test_block_config)
        )

        logger.debug("Request args: %s", request_args)

        self._request_args = request_args

        # There is no way using requests to make a prepared request that will
        # not follow redirects, so instead we have to do this. This also means
        # that we can't have the 'pre-request' hook any more because we don't
        # create a prepared request.

        def prepared_request():
            # If there are open files, create a context manager around each so
            # they will be closed at the end of the request.
            with ExitStack() as stack:
                stack.enter_context(_set_cookies_for_request(session, request_args))

                # These are mutually exclusive
                if file_body:
                    # Any headers will have been set in the above function
                    file = stack.enter_context(open(file_body, "rb"))
                    request_args.update(data=file)
                else:
                    files = get_file_arguments(request_args, stack, test_block_config)
                    if files:
                        logger.debug("Sending %d files in request", len(files["files"]))
                        self._request_args.update(files)

                headers = self._request_args.get("headers", {})
                for k, v in headers.items():
                    headers[str(k)] = str(v)

                return session.request(**self._request_args)

        self._prepared: Callable[[], requests.Response] = prepared_request

    def run(self) -> requests.Response:
        """Runs the prepared request and times it

        Returns:
            response object
        """

        attach_yaml(
            self._request_args,
            name="rest_request",
        )

        try:
            return self._prepared()
        except requests.exceptions.RequestException as e:
            logger.exception("Error running prepared request")
            raise exceptions.RestRequestException from e

    @property
    def request_vars(self) -> Box:
        return Box(self._request_args)



import logging
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Optional, TypeVar, Union

import pytest

from tavern._core.dict_util import format_keys, get_tavern_box
from tavern._core.general import load_global_config
from tavern._core.pytest.config import TavernInternalConfig, TestConfig
from tavern._core.strict_util import StrictLevel

logger: logging.Logger = logging.getLogger(__name__)


def add_parser_options(parser_addoption, with_defaults: bool = True) -> None:
    """Add argparse options

    This is shared between the CLI and pytest (for now)

    See also _core.pytesthook.hooks.pytest_addoption
    """
    parser_addoption(
        "--tavern-global-cfg",
        help="One or more global configuration files to include in every test",
        nargs="+",
    )
    parser_addoption(
        "--tavern-http-backend",
        help="Which http backend to use",
        default="requests" if with_defaults else None,
    )
    parser_addoption(
        "--tavern-mqtt-backend",
        help="Which mqtt backend to use",
        default="paho-mqtt" if with_defaults else None,
    )
    parser_addoption(
        "--tavern-grpc-backend",
        help="Which grpc backend to use",
        default="grpc" if with_defaults else None,
    )
    parser_addoption(
        "--tavern-strict",
        help="Default response matching strictness",
        default=None,
        nargs="+",
    )
    parser_addoption(
        "--tavern-use-default-traceback",
        help="Use normal python-style traceback",
        default=False,
        action="store_true",
    )
    parser_addoption(
        "--tavern-always-follow-redirects",
        help="Always follow HTTP redirects",
        default=False,
        action="store_true",
    )
    parser_addoption(
        "--tavern-file-path-regex",
        help="Regex to search for Tavern YAML test files",
        default=r".+\.tavern\.ya?ml$",
        action="store",
        nargs=1,
    )
    parser_addoption(
        "--tavern-setup-init-logging",
        help="Set up a simple logger for tavern initialisation. Only for internal use and debugging, may be removed in future with no warning.",
        default=False,
        action="store_true",
    )


def add_ini_options(parser: pytest.Parser) -> None:
    """Add an option to pass in a global config file for tavern

    See also _core.pytesthook._core.util.add_parser_options
    """
    parser.addini(
        "tavern-global-cfg",
        help="One or more global configuration files to include in every test",
        type="linelist",
        default=[],
    )
    parser.addini(
        "tavern-http-backend", help="Which http backend to use", default="requests"
    )
    parser.addini(
        "tavern-mqtt-backend", help="Which mqtt backend to use", default="paho-mqtt"
    )
    parser.addini(
        "tavern-grpc-backend", help="Which grpc backend to use", default="grpc"
    )
    parser.addini(
        "tavern-strict",
        help="Default response matching strictness",
        type="args",
        default=None,
    )
    parser.addini(
        "tavern-use-default-traceback",
        help="Use normal python-style traceback",
        type="bool",
        default=False,
    )
    parser.addini(
        "tavern-always-follow-redirects",
        help="Always follow HTTP redirects",
        type="bool",
        default=False,
    )
    parser.addini(
        "tavern-file-path-regex",
        help="Regex to search for Tavern YAML test files",
        default=r".+\.tavern\.ya?ml$",
        type="args",
    )
    parser.addini(
        "tavern-setup-init-logging",
        help="Set up a simple logger for tavern initialisation. Only for internal use and debugging, may be removed in future with no warning.",
        type="bool",
        default=False,
    )


def load_global_cfg(pytest_config: pytest.Config) -> TestConfig:
    return _load_global_cfg(pytest_config).with_new_variables()


@lru_cache
def _load_global_cfg(pytest_config: pytest.Config) -> TestConfig:
    """Load globally included config files from cmdline/cfg file arguments

    Args:
        pytest_config: Pytest config object

    Returns:
        variables/stages/etc from global config files

    Raises:
        exceptions.UnexpectedKeysError: Invalid settings in one or more config
            files detected
    """

    # Load ini first
    ini_global_cfg_paths = pytest_config.getini("tavern-global-cfg") or []
    # THEN load command line, to allow overwriting of values
    cmdline_global_cfg_paths = pytest_config.getoption("tavern_global_cfg") or []

    all_paths = ini_global_cfg_paths + cmdline_global_cfg_paths
    global_cfg_dict = load_global_config(all_paths)

    variables: Dict = {}
    try:
        loaded_variables = global_cfg_dict["variables"]
    except KeyError:
        logger.debug("Nothing to format in global config files")
    else:
        tavern_box = get_tavern_box()
        variables = format_keys(loaded_variables, tavern_box)

    global_cfg = TestConfig(
        variables=variables,
        strict=_load_global_strictness(pytest_config),
        follow_redirects=_load_global_follow_redirects(pytest_config),
        tavern_internal=TavernInternalConfig(
            pytest_hook_caller=pytest_config.hook,
            backends=_load_global_backends(pytest_config),
        ),
        stages=global_cfg_dict.get("stages", []),
    )

    return global_cfg


def _load_global_backends(pytest_config: pytest.Config) -> Dict[str, Any]:
    """Load which backend should be used"""
    return {
        b: get_option_generic(pytest_config, f"tavern-{b}-backend", None)
        for b in TestConfig.backends()
    }


def _load_global_strictness(pytest_config: pytest.Config) -> StrictLevel:
    """Load the global 'strictness' setting"""

    options: List = get_option_generic(pytest_config, "tavern-strict", [])

    return StrictLevel.from_options(options)


def _load_global_follow_redirects(pytest_config: pytest.Config) -> bool:
    """Load the global 'follow redirects' setting"""
    return get_option_generic(pytest_config, "tavern-always-follow-redirects", False)


T = TypeVar("T", bound=Optional[Union[str, List, List[Path], List[str], bool]])


def get_option_generic(
    pytest_config: pytest.Config,
    flag: str,
    default: T,
) -> T:
    """Get a configuration option or return the default

    Priority order is cmdline, then ini, then default"""
    cli_flag = flag.replace("-", "_")
    ini_flag = flag

    # Lowest priority
    use = default

    # Middle priority
    if pytest_config.getini(ini_flag) is not None:
        use = pytest_config.getini(ini_flag)

    # Top priority
    if pytest_config.getoption(cli_flag) is not None:
        use = pytest_config.getoption(cli_flag)

    return use



import os
import pathlib
from dataclasses import dataclass
from unittest.mock import Mock, patch

import pytest
from faker import Faker

from tavern._core import exceptions
from tavern._core.pytest.file import YamlFile, _get_parametrized_items


@dataclass
class MockArgs:
    session: pytest.Session
    parent: pytest.File
    path: pathlib.Path


def mock_args():
    """Get a basic test config to initialise a YamlFile object with"""

    path = pathlib.Path("abc")

    cargs = {"rootdir": "abc", "path": path}

    config = Mock(**cargs, rootpath="abc")

    session = Mock(_initialpaths=[], config=config)

    parent = Mock(
        spec=os.PathLike,
        config=config,
        parent=None,
        nodeid="sdlfs",
        **cargs,
        session=session,
    )

    return MockArgs(session, parent, path)


def get_basic_parametrize_mark(faker):
    """Get a random 'normal' parametrize mark"""
    return {"parametrize": {"key": faker.name(), "vals": [faker.name(), 2, 3]}}


def get_joined_parametrize_mark(faker):
    """Get a random 'combined' parametrize mark"""
    return {
        "parametrize": {
            "key": [faker.name(), faker.name()],
            "vals": [["w", "x"], ["y", "z"]],
        }
    }


def get_parametrised_tests(marks):
    args = mock_args()
    y = YamlFile.from_parent(args.parent, path=args.path)
    y.session = args.session

    spec = {"test_name": "a test", "stages": []}

    gen = _get_parametrized_items(y, spec, marks, [])

    return list(gen)


def test_none():
    marks = []

    tests = get_parametrised_tests(marks)

    # Only 1
    assert len(tests) == 1


@pytest.mark.parametrize("faker", [Faker(), Faker("zh_CN")])
class TestMakeFile:
    def test_only_single(self, faker):
        marks = [get_basic_parametrize_mark(faker)]

        tests = get_parametrised_tests(marks)

        # [1]
        # [2]
        # [3]
        assert len(tests) == 3

    def test_only_basic(self, faker):
        marks = [
            get_basic_parametrize_mark(faker),
            get_basic_parametrize_mark(faker),
            get_basic_parametrize_mark(faker),
        ]
        tests = get_parametrised_tests(marks)

        # [1, 1, 1]
        # [1, 1, 2]
        # [1, 1, 3]
        # [1, 2, 1]
        # [1, 2, 2]
        # [1, 2, 3]
        # etc.
        assert len(tests) == 27

    def test_double(self, faker):
        marks = [get_joined_parametrize_mark(faker), get_basic_parametrize_mark(faker)]

        tests = get_parametrised_tests(marks)

        # [w, x, 1]
        # [w, x, 2]
        # [w, x, 3]
        # [y, z, 1]
        # [y, z, 2]
        # [y, z, 3]
        assert len(tests) == 6

    def test_double_double(self, faker):
        marks = [
            get_joined_parametrize_mark(faker),
            get_joined_parametrize_mark(faker),
            get_basic_parametrize_mark(faker),
        ]

        tests = get_parametrised_tests(marks)

        # [w, x, w, x, 1]
        # [w, x, w, x, 2]
        # [w, x, w, x, 3]
        # [w, x, y, z, 1]
        # [w, x, y, z, 2]
        # [w, x, y, z, 3]
        # etc.
        assert len(tests) == 12

    def test_double_double_single(self, faker):
        marks = [
            get_joined_parametrize_mark(faker),
            get_joined_parametrize_mark(faker),
            get_basic_parametrize_mark(faker),
            get_basic_parametrize_mark(faker),
        ]

        tests = get_parametrised_tests(marks)

        # [w, x, w, x, 1, 1]
        # [w, x, w, x, 1, 2]
        # [w, x, w, x, 2, 1]
        # [w, x, w, x, 2, 2]
        # [w, x, y, z, 1, 1]
        # [w, x, y, z, 1, 2]
        # etc.
        assert len(tests) == 36

    @pytest.mark.parametrize(
        ("keys", "values"),
        (
            ("a", ["b", "c", "d"]),
            (["a"], ["b", "c", "d"]),
            ("a", {"k": "v"}),
            (["a"], {"k": "v"}),
            (["a", "b"], [["b", "c"]]),
            (["a", "b"], [["b", "c"], [{"a": "b"}, {"a": "b"}]]),
            (["a", "b"], [["b", "c"], ["b", "c"], ["d", "e"]]),
        ),
    )
    def test_ext_function_top_level(self, faker, keys, values):
        with patch(
            "tavern._core.pytest.file.get_wrapped_create_function",
            lambda _: lambda: values,
        ):
            marks = [
                {"parametrize": {"key": keys, "vals": {"$ext": {"function": "a:v"}}}}
            ]

            tests = get_parametrised_tests(marks)

            assert len(tests) == len(values)

    @pytest.mark.parametrize(
        ("keys", "values"),
        (
            # must return a list of lists
            (["a", "b"], {"a": "b"}),
            # must return a list of lists
            (["a", "b"], [{"a": "b"}]),
            # must return a list of lists
            (["a", "b"], [{"a": "b"}, {"a": "b"}]),
            # must return a list of lists
            (["a", "b"], "b"),
            # must return a list of lists
            (["a", "b"], ["b", "c"]),
            # must return a list of lists, where each element is also 3 long
            (["a", "b"], [["b", "c", "e"]]),
            # must return a list of lists, where each element is also 3 long
            (["a", "b"], [["b"]]),
        ),
    )
    def test_ext_function_top_level_invalid(self, faker, keys, values):
        with patch(
            "tavern._core.pytest.file.get_wrapped_create_function",
            lambda _: lambda: values,
        ):
            marks = [
                {"parametrize": {"key": keys, "vals": {"$ext": {"function": "a:v"}}}}
            ]

            with pytest.raises(exceptions.BadSchemaError):
                get_parametrised_tests(marks)


def test_doc_string():
    args = mock_args()
    y = YamlFile.from_parent(args.parent, path=args.path)

    assert isinstance(y.obj.__doc__, str)



"""VERY simple skeleton for plugin stuff

This is here mainly to make MQTT easier, this will almost defintiely change
significantly if/when a proper plugin system is implemented!
"""

import dataclasses
import logging
from functools import partial
from typing import Any, Callable, Dict, List, Mapping, Optional, Protocol, Type

import stevedore

from tavern._core import exceptions
from tavern._core.dict_util import format_keys
from tavern._core.pytest.config import TestConfig
from tavern.request import BaseRequest
from tavern.response import BaseResponse

logger: logging.Logger = logging.getLogger(__name__)


class PluginHelperBase:
    """Base for plugins"""


def plugin_load_error(mgr, entry_point, err):
    """Handle import errors"""
    msg = f"Error loading plugin {entry_point} - {err}"
    raise exceptions.PluginLoadError(msg) from err


class _TavernPlugin(Protocol):
    """A tavern plugin"""

    session_type: Type[Any]
    request_type: Type[BaseRequest]
    verifier_type: Type[BaseResponse]
    response_block_name: str
    request_block_name: str
    schema: Mapping

    def get_expected_from_request(
        self, response_block: BaseResponse, test_block_config: TestConfig, session: Any
    ) -> Any: ...


def is_valid_reqresp_plugin(ext: stevedore.extension.Extension) -> bool:
    """Whether this is a valid 'reqresp' plugin

    Requires certain functions/variables to be present

    Todo:
        Not all of these are required for all request/response types probably

    Args:
        ext: class or module plugin object

    Returns:
        Whether the plugin has everything we need to use it
    """
    required = [
        # MQTTClient, requests.Session
        "session_type",
        # RestRequest, MQTTRequest
        "request_type",
        # request, mqtt_publish, grpc_request
        "request_block_name",
        # Some function that returns a dict
        "get_expected_from_request",
        # MQTTResponse, RestResponse
        "verifier_type",
        # response, mqtt_response, grpc_request
        "response_block_name",
        # dictionary with pykwalify schema
        "schema",
    ]

    plugin: _TavernPlugin = ext.plugin

    return all(hasattr(plugin, i) for i in required)


class _Plugin:
    """Wrapped tavern plugin for convenience"""

    name: str
    plugin: _TavernPlugin


@dataclasses.dataclass
class _PluginCache:
    plugins: List[_Plugin] = dataclasses.field(default_factory=list)

    def __call__(self, config: Optional[TestConfig] = None) -> List[_Plugin]:
        if self.plugins:
            return self.plugins

        if config:
            # NOTE: This is reloaded every time
            self.plugins = self._load_plugins(config)
            return self.plugins

        raise exceptions.PluginLoadError("No config to load plugins from")

    def _load_plugins(self, test_block_config: TestConfig) -> List[_Plugin]:
        """Load plugins from the 'tavern' entrypoint namespace

        This can be a module or a class as long as it defines the right things

        Todo:
            - Limit which plugins are loaded based on some config/command line
              option
            - Different plugin names

        Args:
            test_block_config: available config for test

        Raises:
            exceptions.MissingSettingsError: invalid entry points set

        Returns:
            Loaded plugins, can be a class or a module
        """

        plugins = []

        def enabled(current_backend, ext):
            return (
                ext.name == test_block_config.tavern_internal.backends[current_backend]
            )

        for backend in test_block_config.backends():
            logger.debug("loading backend for %s", backend)

            namespace = f"tavern_{backend}"

            manager = stevedore.EnabledExtensionManager(
                namespace=namespace,
                check_func=partial(enabled, backend),
                verify_requirements=True,
                on_load_failure_callback=plugin_load_error,
            )
            manager.propagate_map_exceptions = True

            manager.map(is_valid_reqresp_plugin)

            if len(manager.extensions) != 1:
                raise exceptions.MissingSettingsError(
                    "Expected exactly one entrypoint in 'tavern-{}' namespace but got {}".format(
                        backend, len(manager.extensions)
                    )
                )

            plugins.extend(manager.extensions)

        return plugins


load_plugins = _PluginCache()


def get_extra_sessions(test_spec: Mapping, test_block_config: TestConfig) -> dict:
    """Get extra 'sessions' for any extra test types

    Args:
        test_spec: Spec for the test block
        test_block_config: available config for test

    Returns:
        mapping of name to session. Session should be a context manager.
    """

    sessions = {}

    plugins = load_plugins(test_block_config)

    for p in plugins:
        if any(
            (p.plugin.request_block_name in i or p.plugin.response_block_name in i)
            for i in test_spec["stages"]
        ):
            logger.debug(
                "Initialising session for %s (%s)", p.name, p.plugin.session_type
            )
            session_spec = test_spec.get(p.name, {})
            formatted = format_keys(session_spec, test_block_config.variables)
            sessions[p.name] = p.plugin.session_type(**formatted)

    return sessions


def get_request_type(
    stage: Mapping,
    test_block_config: TestConfig,
    sessions: Mapping,
) -> BaseRequest:
    """Get the request object for this stage

    there can only be one

    Args:
        stage: spec for this stage
        test_block_config: variables for this test run
        sessions: all available sessions

    Returns:
        request object with a run() method

    Raises:
        exceptions.DuplicateKeysError: More than one kind of request specified
        exceptions.MissingKeysError: No request type specified
    """

    plugins = load_plugins(test_block_config)

    keys = {}

    for p in plugins:
        keys[p.plugin.request_block_name] = p.plugin.request_type

    if len(set(keys) & set(stage)) > 1:
        raise exceptions.DuplicateKeysError(
            f"Can only specify 1 request type but got {set(keys)}"
        )
    elif not list(set(keys) & set(stage)):
        raise exceptions.MissingKeysError(
            f"Need to specify one of valid request types: '{set(keys.keys())}'"
        )

    # We've validated that 1 and only 1 is there, so just loop until the first
    # one is found
    for p in plugins:
        try:
            request_args = stage[p.plugin.request_block_name]
        except KeyError:
            pass
        else:
            session = sessions[p.name]
            request_class = p.plugin.request_type
            logger.debug(
                "Initialising request class for %s (%s)", p.name, request_class
            )
            break

    request_maker = request_class(session, request_args, test_block_config)

    return request_maker


class ResponseVerifier(dict):
    plugin_name: str


def _foreach_response(
    stage: Mapping,
    test_block_config: TestConfig,
    action: Callable[[_Plugin, str], dict],
) -> Dict[str, dict]:
    """Do something for each response

    Args:
        stage: Stage of test
        test_block_config: Config for test
        action: function that takes (plugin, response block)

    Returns:
        mapping of plugin name to list of expected (normally length 1)
    """

    plugins = load_plugins(test_block_config)

    retvals = {}

    for p in plugins:
        response_block = stage.get(p.plugin.response_block_name)
        if response_block is not None:
            retvals[p.name] = action(p, response_block)

    return retvals


def get_expected(
    stage: Mapping,
    test_block_config: TestConfig,
    sessions: Mapping,
):
    """Get expected responses for each type of request

    Though only 1 request can be made, it can cause multiple responses.

    Because we need to subcribe to MQTT topics, which might be formatted from
    keys from included files, the 'expected'/'response' needs to be formatted
    BEFORE running the request.

    Args:
        stage: test stage
        test_block_config: available configuration for this test
        sessions: all available sessions

    Returns:
        mapping of request type to expected response dict
    """

    def action(p, response_block):
        plugin_expected = p.plugin.get_expected_from_request(
            response_block, test_block_config, sessions[p.name]
        )
        if plugin_expected:
            plugin_expected = ResponseVerifier(**plugin_expected)
            plugin_expected.plugin_name = p.name
            return plugin_expected
        else:
            return None

    return _foreach_response(stage, test_block_config, action)


def get_verifiers(
    stage: Mapping,
    test_block_config: TestConfig,
    sessions: Mapping,
    expected: Mapping,
):
    """Get one or more response validators for this stage

    Args:
        stage: spec for this stage
        test_block_config: variables for this test run
        sessions: all available sessions
        expected: expected responses for this stage

    Returns:
        response validator object with a verify(response) method
    """

    def action(p, _):
        session = sessions[p.name]
        logger.debug(
            "Initialising verifier for %s (%s)", p.name, p.plugin.verifier_type
        )
        verifiers = []

        plugin_expected = expected[p.name]

        verifier = p.plugin.verifier_type(
            session, stage["name"], plugin_expected, test_block_config
        )
        verifiers.append(verifier)

        return verifiers

    return _foreach_response(stage, test_block_config, action)



import operator
import re
from typing import Any, Dict, List, Sized

from tavern._core import exceptions


def test_type(val, mytype) -> bool:
    """Check value fits one of the types, if so return true, else false"""
    typelist = TYPES.get(str(mytype).lower())
    if typelist is None:
        raise TypeError(
            f"Type {str(mytype).lower()} is not a valid type to test against!"
        )
    try:
        for testtype in typelist:
            if isinstance(val, testtype):  # type: ignore
                return True
        return False
    except TypeError:
        return isinstance(val, typelist)  # type: ignore


COMPARATORS = {
    "count_eq": lambda x, y: safe_length(x) == y,
    "lt": operator.lt,
    "less_than": operator.lt,
    "eq": operator.eq,
    "equals": operator.eq,
    "str_eq": lambda x, y: operator.eq(str(x), str(y)),
    "ne": operator.ne,
    "not_equals": operator.ne,
    "gt": operator.gt,
    "greater_than": operator.gt,
    "contains": lambda x, y: x and operator.contains(x, y),  # is y in x
    "contained_by": lambda x, y: y and operator.contains(y, x),  # is x in y
    "regex": lambda x, y: regex_compare(str(x), str(y)),
    "type": test_type,
}
TYPES: Dict[str, List[Any]] = {
    "none": [type(None)],
    "number": [int, float],
    "int": [int],
    "float": [float],
    "bool": [bool],
    "str": [str],
    "list": [list],
    "dict": [dict],
}


def regex_compare(_input, regex) -> bool:
    return bool(re.search(regex, _input))


def safe_length(var: Sized) -> int:
    """Exception-safe length check, returns -1 if no length on type or error"""
    try:
        return len(var)
    except TypeError:
        return -1


def validate_comparison(each_comparison: Dict[Any, Any]):
    if extra := set(each_comparison.keys()) - {"jmespath", "operator", "expected"}:
        raise exceptions.BadSchemaError(
            "Invalid keys given to JMES validation function (got extra keys: {})".format(
                extra
            )
        )

    jmespath, _operator, expected = (
        each_comparison["jmespath"],
        each_comparison["operator"],
        each_comparison["expected"],
    )

    try:
        COMPARATORS[_operator]
    except KeyError as e:
        raise exceptions.BadSchemaError("Invalid comparator given") from e

    return jmespath, _operator, expected


def actual_validation(
    _operator: str, _actual, expected, _expression, expression
) -> None:
    if not COMPARATORS[_operator](_actual, expected):
        raise exceptions.JMESError(f"Validation '{expression}' ({_expression}) failed!")



import copy
from unittest.mock import Mock

import pytest

from tavern._core.plugins import load_plugins
from tavern._core.pytest.config import TavernInternalConfig, TestConfig
from tavern._core.strict_util import StrictLevel

_includes = TestConfig(
    variables={
        "request": {"prefix": "www.", "url": "google.com"},
        "test_auth_token": "abc123",
        "code": "def456",
        "callback_url": "www.yahoo.co.uk",
        "request_topic": "/abc",
    },
    strict=StrictLevel.all_on(),
    tavern_internal=TavernInternalConfig(
        pytest_hook_caller=Mock(),
        backends={"mqtt": "paho-mqtt", "http": "requests", "grpc": "grpc"},
    ),
    follow_redirects=False,
    stages=[],
)


@pytest.fixture(scope="function", name="includes")
def fix_example_includes():
    return copy.deepcopy(_includes)


@pytest.fixture(scope="session", autouse=True)
def initialise_plugins():
    load_plugins(_includes)



import dataclasses
import os
import tempfile
from contextlib import ExitStack
from textwrap import dedent
from unittest.mock import Mock

import pytest
import requests
import yaml
from requests.cookies import RequestsCookieJar

from tavern._core import exceptions
from tavern._core.extfunctions import update_from_ext
from tavern._plugins.rest.files import get_file_arguments
from tavern._plugins.rest.request import (
    RestRequest,
    _check_allow_redirects,
    _read_expected_cookies,
    get_request_args,
)


@pytest.fixture(name="req")
def fix_example_request():
    spec = {
        "url": "{request.prefix:s}{request.url:s}",
        "method": "POST",
        "headers": {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {test_auth_token:s}",
        },
        "data": {
            "a_thing": "authorization_code",
            "code": "{code:s}",
            "url": "{callback_url:s}",
            "array": ["{code:s}", "{code:s}"],
        },
    }

    return spec.copy()


class TestRequests:
    def test_unknown_fields(self, req, includes):
        """Unkown args should raise an error"""
        req["fodokfowe"] = "Hello"

        with pytest.raises(exceptions.UnexpectedKeysError):
            RestRequest(Mock(), req, includes)

    def test_missing_format(self, req, includes):
        """All format variables should be present"""
        del includes.variables["code"]

        with pytest.raises(exceptions.MissingFormatError):
            RestRequest(Mock(), req, includes)

    def test_bad_get_body(self, req, includes):
        """Can't add a body with a GET request"""
        req["method"] = "GET"

        with pytest.warns(RuntimeWarning):
            RestRequest(
                Mock(spec=requests.Session, cookies=RequestsCookieJar()), req, includes
            )


class TestHttpRedirects:
    def test_session_called_no_redirects(self, req, includes):
        """Always disable redirects by defauly"""

        assert _check_allow_redirects(req, includes) is False

    @pytest.mark.parametrize("do_follow", [True, False])
    def test_session_do_follow_redirects_based_on_test(self, req, includes, do_follow):
        """Locally enable following redirects in test"""

        req["follow_redirects"] = do_follow

        assert _check_allow_redirects(req, includes) == do_follow

    @pytest.mark.parametrize("do_follow", [True, False])
    def test_session_do_follow_redirects_based_on_global_flag(
        self, req, includes, do_follow
    ):
        """Globally enable following redirects in test"""

        includes = dataclasses.replace(includes, follow_redirects=do_follow)

        assert _check_allow_redirects(req, includes) == do_follow


class TestCookies:
    @pytest.fixture
    def mock_session(self):
        return Mock(spec=requests.Session, cookies=RequestsCookieJar())

    def test_no_expected_none_available(self, mock_session, req, includes):
        """No cookies expected and none available = OK"""

        req["cookies"] = []

        assert _read_expected_cookies(mock_session, req, includes) == {}

    def test_available_not_waited(self, req, includes):
        """some available but not set"""

        cookiejar = RequestsCookieJar()
        cookiejar.set("a", 2)
        mock_session = Mock(spec=requests.Session, cookies=cookiejar)

        assert _read_expected_cookies(mock_session, req, includes) is None

    def test_ask_for_nothing(self, req, includes):
        """explicitly ask fo rno cookies"""

        cookiejar = RequestsCookieJar()
        cookiejar.set("a", 2)
        mock_session = Mock(spec=requests.Session, cookies=cookiejar)

        req["cookies"] = []

        assert _read_expected_cookies(mock_session, req, includes) == {}

    def test_not_available_but_wanted(self, mock_session, req, includes):
        """Some wanted but not available"""

        req["cookies"] = ["a"]

        with pytest.raises(exceptions.MissingCookieError):
            _read_expected_cookies(mock_session, req, includes)

    def test_available_and_waited(self, req, includes):
        """some available and wanted"""

        cookiejar = RequestsCookieJar()
        cookiejar.set("a", 2)

        req["cookies"] = ["a"]

        mock_session = Mock(spec=requests.Session, cookies=cookiejar)

        assert _read_expected_cookies(mock_session, req, includes) == {"a": 2}

    def test_format_cookies(self, req, includes):
        """cookies in request should be formatted"""

        cookiejar = RequestsCookieJar()
        cookiejar.set("a", 2)

        req["cookies"] = ["{cookiename}"]
        includes.variables["cookiename"] = "a"

        mock_session = Mock(spec=requests.Session, cookies=cookiejar)

        assert _read_expected_cookies(mock_session, req, includes) == {"a": 2}

    def test_no_overwrite_cookie(self, req, includes):
        """cant redefine a cookie from previous request"""

        cookiejar = RequestsCookieJar()
        cookiejar.set("a", 2)

        req["cookies"] = ["a", {"a": "sjidfsd"}]

        mock_session = Mock(spec=requests.Session, cookies=cookiejar)

        with pytest.raises(exceptions.DuplicateCookieError):
            _read_expected_cookies(mock_session, req, includes)

    def test_no_duplicate_cookie(self, req, includes):
        """Can't override a cookiev alue twice"""

        cookiejar = RequestsCookieJar()

        req["cookies"] = [{"a": "sjidfsd"}, {"a": "fjhj"}]

        mock_session = Mock(spec=requests.Session, cookies=cookiejar)

        with pytest.raises(exceptions.DuplicateCookieError):
            _read_expected_cookies(mock_session, req, includes)


class TestRequestArgs:
    def test_default_method(self, req, includes):
        del req["method"]
        del req["data"]

        args = get_request_args(req, includes)

        assert args["method"] == "GET"

    @pytest.mark.parametrize("body_key", ("json", "data"))
    def test_default_method_raises_with_body(self, req, includes, body_key):
        del req["method"]
        del req["data"]

        req[body_key] = {"a": "b"}

        with pytest.warns(RuntimeWarning):
            get_request_args(req, includes)

    def test_no_override_method(self, req, includes):
        req["method"] = "POST"

        args = get_request_args(req, includes)

        assert args["method"] == "POST"

    @pytest.mark.parametrize("extra", [{}, {"json": [1, 2, 3]}, {"data": {"a": 2}}])
    def test_no_default_content_type(self, req, includes, extra):
        del req["headers"]["Content-Type"]
        req.pop("json", {})
        req.pop("data", {})

        req.update(**extra)

        args = get_request_args(req, includes)

        # Requests will automatically set content type headers for json/form encoded data so we don't need to
        with pytest.raises(KeyError):
            assert args["headers"]["content-type"]

    def test_no_set_content_type(self, req, includes):
        del req["headers"]["Content-Type"]

        args = get_request_args(req, includes)

        with pytest.raises(KeyError):
            assert args["headers"]["content-type"]

    def test_cannot_send_data_and_json(self, req, includes):
        req["json"] = [1, 2, 3]
        req["data"] = [1, 2, 3]

        with pytest.raises(exceptions.BadSchemaError):
            get_request_args(req, includes)

    def test_no_override_content_type(self, req, includes):
        req["headers"]["Content-Type"] = "application/x-www-form-urlencoded"

        args = get_request_args(req, includes)

        assert args["headers"]["Content-Type"] == "application/x-www-form-urlencoded"

    def test_no_override_content_type_case_insensitive(self, req, includes):
        del req["headers"]["Content-Type"]
        req["headers"]["content-type"] = "application/x-www-form-urlencoded"

        args = get_request_args(req, includes)

        assert args["headers"]["content-type"] == "application/x-www-form-urlencoded"

    def test_nested_params_encoded(self, req, includes):
        req["params"] = {"a": {"b": {"c": "d"}}}

        args = get_request_args(req, includes)

        assert args["params"]["a"] == "%7B%22b%22%3A+%7B%22c%22%3A+%22d%22%7D%7D"

    def test_array_substitution(self, req, includes):
        args = get_request_args(req, includes)

        assert args["data"]["array"] == ["def456", "def456"]

    def test_file_and_json_fails(self, req, includes):
        """Can't send json and files at once"""
        req["files"] = ["abc"]
        req["json"] = {"key": "value"}

        with pytest.raises(exceptions.BadSchemaError):
            get_request_args(req, includes)

    def test_file_and_data_succeeds(self, req, includes):
        """Can send form data and files at once"""
        req["files"] = ["abc"]

        get_request_args(req, includes)

    @pytest.mark.parametrize("extra_headers", ({}, {"x-cool-header": "plark"}))
    def test_headers_no_content_type_change(self, req, includes, extra_headers):
        """Sending a file doesn't set the content type as json"""
        del req["data"]
        req["files"] = ["abc"]

        args = get_request_args(req, includes)

        assert "content-type" not in [i.lower() for i in args["headers"].keys()]

    @pytest.mark.parametrize("cert_value", ("a", ("a", "b"), ["a", "b"]))
    def test_cert_with_valid_values(self, req, includes, cert_value):
        req["cert"] = cert_value
        args = get_request_args(req, includes)
        if isinstance(cert_value, list):
            assert args["cert"] == (cert_value[0], cert_value[1])
        else:
            assert args["cert"] == cert_value

    @pytest.mark.parametrize("verify_values", (True, False, "a"))
    def test_verity_with_valid_values(self, req, includes, verify_values):
        req["verify"] = verify_values
        args = get_request_args(req, includes)

        assert args["verify"] == verify_values


class TestExtFunctions:
    def test_get_from_function(self, req, includes):
        """Make sure ext functions work in request

        This is a bit of a silly example because we're passing a dictionary
        instead of a string like it would be from the test, but it saves us
        having to define another external function just for this test
        """
        to_copy = {"thing": "value"}
        original_json = {"test": "test"}

        req["json"] = {
            "$ext": {"function": "copy:copy", "extra_args": [to_copy]},
            **original_json,
        }

        update_from_ext(req, ["json"])

        assert req["json"] == dict(**to_copy, **original_json)


class TestOptionalDefaults:
    @pytest.mark.parametrize("verify", (True, False))
    def test_passthrough_verify(self, req, includes, verify):
        """Should be able to pass 'verify' through to requests.request"""

        req["verify"] = verify

        args = get_request_args(req, includes)

        assert args["verify"] == verify


class TestFileBody:
    def test_file_body_format(self, req, includes):
        """Test getting file body"""

        req.pop("data")

        with tempfile.NamedTemporaryFile(encoding="utf8", mode="w") as tmpin:
            tmpin.write("OK")
            includes.variables["tmpfile_loc"] = tmpin.name

            req["file_body"] = "{tmpfile_loc}"

            args = get_request_args(req, includes)

        assert args["file_body"] == tmpin.name

    def test_file_body_content_type(self, req, includes):
        """Test inferring content type etc. works"""

        req.pop("data")
        req.pop("headers")

        with tempfile.NamedTemporaryFile(
            encoding="utf8", mode="w", suffix=".json"
        ) as tmpin:
            tmpin.write("OK")

            req["file_body"] = tmpin.name

            args = get_request_args(req, includes)

        assert args["file_body"] == tmpin.name
        assert args["headers"]["content-type"] == "application/json"

    def test_file_body_content_encoding(self, req, includes):
        """Test inferring content type etc. works"""

        req.pop("data")
        req.pop("headers")

        with tempfile.NamedTemporaryFile(
            encoding="utf8", mode="w", suffix=".tar.gz"
        ) as tmpin:
            tmpin.write("OK")

            req["file_body"] = tmpin.name

            args = get_request_args(req, includes)

        assert args["file_body"] == tmpin.name
        assert args["headers"]["content-type"] == "application/x-tar"
        assert args["headers"]["Content-Encoding"] == "gzip"


class TestGetFiles:
    @pytest.fixture
    def mock_stack(self):
        return Mock(spec=ExitStack)

    def test_get_no_files(self, mock_stack, includes):
        """No files in request -> no files"""

        request_args = {}

        assert get_file_arguments(request_args, mock_stack, includes) == {}

    def test_get_empty_files_list(self, mock_stack, includes):
        """No specific files specified -> no files"""

        request_args = {"files": {}}

        assert get_file_arguments(request_args, mock_stack, includes) == {}

    def test_a_file(self, mock_stack, includes):
        """Json file should have the correct mimetype etc."""

        with tempfile.NamedTemporaryFile(suffix=".json") as tfile:
            request_args = {"files": {"file1": tfile.name}}

            file_spec = get_file_arguments(request_args, mock_stack, includes)

        file = file_spec["files"]["file1"]
        assert file[0] == os.path.basename(tfile.name)
        assert file[2] == "application/json"

    def test_use_long_form_content_type(self, mock_stack, includes):
        """Use custom content type"""

        with tempfile.NamedTemporaryFile(suffix=".json") as tfile:
            request_args = {
                "files": {
                    "file1": {
                        "file_path": tfile.name,
                        "content_type": "abc123",
                        "content_encoding": "def456",
                    }
                }
            }

            file_spec = get_file_arguments(request_args, mock_stack, includes)

        file = file_spec["files"]["file1"]
        assert file[0] == os.path.basename(tfile.name)
        assert file[2] == "abc123"
        assert file[3] == {"Content-Encoding": "def456"}

    @pytest.mark.parametrize(
        "file_args",
        [
            {
                "file1": {
                    "file_path": "{tmpname}",
                    "content_type": "abc123",
                    "content_encoding": "def456",
                }
            },
            {"file1": "{tmpname}"},
        ],
    )
    def test_format_filename(self, mock_stack, includes, file_args):
        """Filenames should be formatted in short and long styles"""

        with tempfile.NamedTemporaryFile(suffix=".json") as tfile:
            includes.variables["tmpname"] = tfile.name
            request_args = {"files": {"file1": tfile.name}}

            file_spec = get_file_arguments(request_args, mock_stack, includes)

        file = file_spec["files"]["file1"]
        assert file[0] == os.path.basename(tfile.name)

    def test_grouped_file_names(self, mock_stack, includes):
        """Parse grouped names appropriately"""
        with tempfile.NamedTemporaryFile() as tfile:
            raw_yaml_args = """
                # Send file_1.txt and file_2.txt, both with name="input_files", in the multipart data.
                - form_field_name: "input_files"
                  file_path: "%FILENAME%"
                  content_type: "application/customtype"
                  content_encoding: "UTF16"
                - form_field_name: "input_files"
                  file_path: "%FILENAME%"
                  content_type: "application/json"
                """

            raw_yaml_args = raw_yaml_args.replace("%FILENAME%", tfile.name)

            file_args = yaml.safe_load(dedent(raw_yaml_args))

            request_args = {"files": file_args}

            parsed = get_file_arguments(request_args, mock_stack, includes)

            parsed_into = {
                "files": [
                    (
                        "input_files",
                        (
                            os.path.basename(tfile.name),
                            mock_stack.enter_context.return_value,
                            "application/customtype",
                            {"Content-Encoding": "UTF16"},
                        ),
                    ),
                    (
                        "input_files",
                        (
                            os.path.basename(tfile.name),
                            mock_stack.enter_context.return_value,
                            "application/json",
                        ),
                    ),
                ],
            }

        assert parsed_into == parsed



import base64
import gzip
import itertools
import json
import math
import mimetypes
import os
import time
import uuid
from datetime import datetime, timedelta
from hashlib import sha512
from urllib.parse import unquote_plus, urlencode

import jwt
from flask import Flask, Response, jsonify, make_response, redirect, request, session
from itsdangerous import URLSafeTimedSerializer

app = Flask(__name__)
app.config.update(SECRET_KEY="secret")


@app.route("/token", methods=["GET"])
def token():
    return (
        '<div><a src="http://127.0.0.1:5003/verify?token=c9bb34ba-131b-11e8-b642-0ed5f89f718b">Link</a></div>',
        200,
    )


@app.route("/headers", methods=["GET"])
def headers():
    return "OK", 200, {"X-Integration-Value": "_HelloWorld1", "ATestHEader": "orange"}


@app.route("/verify", methods=["GET"])
def verify():
    if request.args.get("token") == "c9bb34ba-131b-11e8-b642-0ed5f89f718b":
        return "", 200
    else:
        return "", 401


@app.route("/get_thing_slow", methods=["GET"])
def get_slow():
    time.sleep(0.25)

    response = {"status": "OK"}
    return jsonify(response), 200


@app.route("/fake_dictionary", methods=["GET"])
def get_fake_dictionary():
    fake = {
        "top": {"Thing": "value", "nested": {"doubly": {"inner": "value"}}},
        "an_integer": 123,
        "a_string": "abc",
        "a_bool": True,
    }

    return jsonify(fake), 200


@app.route("/fake_list", methods=["GET"])
def list_response():
    list_response = ["a", "b", "c", 1, 2, 3, -1.0, -2.0, -3.0]
    return jsonify(list_response), 200


@app.route("/complicated_list", methods=["GET"])
def complicated_list():
    list_response = ["a", {"b": "c"}]
    return jsonify(list_response), 200


@app.route("/nested_list", methods=["GET"])
def nested_list_response():
    response = {"top": ["a", "b", {"key": "val"}]}
    return jsonify(response), 200


@app.route("/fake_upload_file", methods=["POST"])
def upload_fake_file():
    if not request.files:
        return "", 401

    return _handle_files()


def _handle_files():
    for item in request.files.values():
        if item.filename:
            filetype = ".{}".format(item.filename.split(".")[-1])
            if filetype in mimetypes.suffix_map:
                if not item.content_type:
                    return "", 400
    # Try to download each of the files downloaded to /tmp and
    # then remove them
    for key in request.files:
        file_to_save = request.files[key]
        path = os.path.join("/tmp", file_to_save.filename)
        file_to_save.save(path)
    return "", 200


class BadFileUploadException(Exception):
    """Something wrong when uploading files"""


def _verify_is_file_multipart():
    if not mimetypes.inited:
        mimetypes.init()

    if not request.content_type.startswith("multipart/form-data"):
        raise BadFileUploadException("Was not a multipart form upload")

    if not request.files:
        raise BadFileUploadException("No files in request")


@app.route("/fake_upload_file_data", methods=["POST"])
def upload_fake_file_and_data():
    try:
        _verify_is_file_multipart()
    except BadFileUploadException as e:
        return jsonify({"error": str(e)}), 400

    if not request.form.to_dict():
        return "", 400

    return _handle_files()


@app.route("/files_expect_in_order", methods=["POST"])
def upload_specific_files_in_order():
    """Expects a multipart form upload with files in the correct order

    See test_files.tavern.yaml for expected list of files here
    """

    try:
        _verify_is_file_multipart()
    except BadFileUploadException as e:
        return jsonify({"error": str(e)}), 400

    try:
        group_1 = request.files.getlist("group_1")
        if len(group_1) != 2:
            raise Exception(f"expected 2 files in group 1, got {len(group_1)}")
        if group_1[0].filename != "OK.txt":
            raise Exception(
                f"First file in group 1 should be OK.txt, was {group_1[0].filename}"
            )
        if group_1[1].filename != "OK.json.gz":
            raise Exception(
                f"Second file in group 1 should be OK.json.gz, was {group_1[1].filename}"
            )

        group_2 = request.files.getlist("group_2")
        if len(group_2) != 1:
            raise Exception(f"expected 1 files in group 2, got {len(group_2)}")
        if group_2[0].filename != "OK.txt":
            raise Exception(
                f"First file in group 2 should be OK.txt, was {group_2[0].filename}"
            )
    except Exception as e:
        return jsonify({"error": str(e)}), 400

    return "", 200


@app.route("/nested/again", methods=["GET"])
def multiple_path_items_response():
    response = {"status": "OK"}
    return jsonify(response), 200


@app.route("/pi", methods=["GET"])
def return_fp_number():
    response = {"pi": math.pi}
    return jsonify(response), 200


@app.route("/expect_dtype", methods=["POST"])
def expect_type():
    body = request.get_json()
    value = body.get("value")
    dtype = body.get("dtype")
    dvalue = body.get("dvalue")

    status = "OK"
    code = 200

    if not value and dtype and dvalue:
        status = "Missing expected type or value"
        code = 400

    if str(type(value)) != f"<class '{dtype}'>":
        status = f"Unexpected type: '{str(type(value))}'"
        code = 400

    if value != dvalue:
        status = f"Unexpected value: '{value}'"
        code = 400

    return jsonify({"status": status}), code


@app.route("/status_code_return", methods=["POST"])
def status_code_return():
    body = request.get_json()
    response = {}
    return jsonify(response), int(body["status_code"])


@app.route("/echo", methods=["POST"])
def echo_values():
    body = request.get_json(silent=True)
    response = body
    return jsonify(response), 200


@app.route("/echo_params", methods=["GET"])
def echo_params():
    params = request.args

    response = {}
    for k, v in params.items():
        unquoted = unquote_plus(v)
        try:
            response[k] = json.loads(unquoted)
        except json.decoder.JSONDecodeError:
            response[k] = unquoted

    return jsonify(response), 200


@app.route("/expect_raw_data", methods=["POST"])
def expect_raw_data():
    raw_data = request.stream.read().decode("utf8").strip()
    if raw_data == "OK":
        response = {"status": "ok"}
        code = 200
    elif raw_data == "DENIED":
        response = {"status": "denied"}
        code = 401
    else:
        response = {"status": f"err: '{raw_data}'"}
        code = 400

    return jsonify(response), code


@app.route("/expect_compressed_data", methods=["POST"])
def expect_compressed_data():
    content_type_header = request.headers.get("content-type")
    if content_type_header != "application/json":
        return jsonify("invalid content type " + content_type_header), 400

    content_encoding_header = request.headers.get("content-encoding")
    if content_encoding_header != "gzip":
        return jsonify("invalid content encoding " + content_encoding_header), 400

    compressed_data = request.stream.read()

    decompressed = gzip.decompress(compressed_data)

    raw_data = decompressed.decode("utf8").strip()

    loaded = json.loads(raw_data)

    if loaded == "OK":
        response = {"status": "ok"}
        code = 200
    else:
        response = {"status": f"err: '{raw_data}'"}
        code = 400

    return jsonify(response), code


@app.route("/form_data", methods=["POST"])
def echo_form_values():
    body = request.get_data()
    key, _, value = body.decode("utf8").partition("=")
    response = {key: value}
    return jsonify(response), 200


@app.route("/stream_file", methods=["GET"])
def stream_file():
    def iter():
        for data in range(1, 10):
            yield bytes(data)

    response = Response(iter(), mimetype="application/octet-stream")
    response.headers["Content-Disposition"] = "attachment; filename=tmp.txt"
    return response


statuses = itertools.cycle(["processing", "ready"])


@app.route("/poll", methods=["GET"])
def poll():
    response = {"status": next(statuses)}
    return jsonify(response)


def _maybe_get_cookie_name():
    return (request.get_json(silent=True) or {}).get("cookie_name", "tavern-cookie")


@app.route("/get_cookie", methods=["POST"])
def give_cookie():
    cookie_name = _maybe_get_cookie_name()
    response = Response()
    response.set_cookie(cookie_name, base64.b64encode(os.urandom(16)).decode("utf8"))
    return response, 200


@app.route("/expect_cookie", methods=["GET"])
def expect_cookie():
    cookie_name = _maybe_get_cookie_name()
    if cookie_name not in request.cookies:
        return (
            jsonify({"error": f"No cookie named {cookie_name} in request"}),
            400,
        )
    else:
        return jsonify({"status": "ok"}), 200


@app.route("/redirect/source", methods=["GET"])
def redirect_to_other_endpoint():
    query_params = urlencode(
        {
            "test_value": "lorem ipsum?",
        }
    )

    return redirect(f"/redirect/destination?{query_params}", 302)


@app.route("/redirect/loop", methods=["GET"])
def redirect_loop():
    try:
        if redirect_loop.tries > 50:
            return redirect("/redirect/destination", 302)
        else:
            redirect_loop.tries += 1
    except AttributeError:
        redirect_loop.tries = 1

    return redirect("/redirect/loop", 302)


@app.route("/redirect/destination", methods=["GET"])
def get_redirected_to_here():
    return jsonify({"status": "successful redirect"}), 200


@app.route("/get_single_json_item", methods=["GET"])
def return_one_item():
    return jsonify("c82bfa63-fd2a-419a-8c06-21cb283fd9f7"), 200


@app.route("/authtest/basic", methods=["GET"])
def expect_basic_auth():
    auth = request.authorization

    if auth is None:
        return jsonify({"status": "No authorisation"}), 403

    if auth.type == "basic":
        if auth.username == "fakeuser" and auth.password == "fakepass":
            return (
                jsonify(
                    {
                        "auth_type": auth.type,
                        "auth_user": auth.username,
                        "auth_pass": auth.password,
                    }
                ),
                200,
            )
        else:
            return jsonify({"error": "Wrong username/password"}), 401
    else:
        return jsonify({"error": "unrecognised auth type"}), 403


@app.route("/jmes/return_empty_paged", methods=["GET"])
def return_empty_paged():
    return jsonify({"pages": 0, "data": []}), 200


@app.route("/jmes/with_dot", methods=["GET"])
def return_with_dot():
    return jsonify({"data.a": "a", "data.b": "b"}), 200


@app.route("/uuid/v4", methods=["GET"])
def get_uuid_v4():
    return jsonify({"uuid": uuid.uuid4()}), 200


@app.route("/707-regression", methods=["GET"])
def get_707():
    return jsonify({"a": 1, "b": {"first": 10, "second": 20}, "c": 2})


users = {"mark": {"password": "password", "regular": "foo", "protected": "bar"}}

serializer = URLSafeTimedSerializer(
    secret_key="secret",
    salt="cookie",
    signer_kwargs={"key_derivation": "hmac", "digest_method": sha512},
)


@app.route("/withsession/login", methods=["POST"])
def login():
    r = request.get_json()
    username = r["username"]
    password = r["password"]

    if password == users[username]["password"]:
        session["user"] = username
        response = make_response("", 200)
        response.set_cookie(
            "remember",
            value=serializer.dumps(username),
            expires=datetime.utcnow() + timedelta(days=30),
            httponly=True,
        )
        return response

    return "", 401


@app.route("/withsession/regular", methods=["GET"])
def regular():
    username = session.get("user")

    if not username:
        remember = request.cookies.get("remember")
        if remember:
            username = serializer.loads(remember, max_age=3600)

    if username:
        return jsonify(regular=users[username]["regular"]), 200

    return "", 401


@app.route("/withsession/protected", methods=["GET"])
def protected():
    username = session.get("user")
    if username:
        return jsonify(protected=users[username]["protected"]), 200
    return "", 401


@app.route("/606-regression-list", methods=["GET"])
def get_606_list():
    return jsonify([])


@app.route("/606-regression-dict", methods=["GET"])
def get_606_dict():
    return jsonify({})


@app.route("/magic-multi-method", methods=["GET", "POST", "DELETE"])
def get_any_method():
    return jsonify({"method": request.method})


@app.route("/get_jwt", methods=["POST"])
def get_jwt():
    secret = "240c8c9c-39b9-426b-9503-3126f96c2eaf"
    audience = "testserver"

    r = request.get_json()

    if r["user"] != "test-user" or r["password"] != "correct-password":
        return jsonify({"error": "Incorrect username/password"}), 401

    payload = {
        "sub": "test-user",
        "aud": audience,
        "exp": datetime.utcnow() + timedelta(hours=1),
    }

    token = jwt.encode(payload, secret, algorithm="HS256")

    return jsonify({"jwt": token})



import contextlib
import json
import logging
from typing import Any, Dict, List, Mapping, Union
from urllib.parse import parse_qs, urlparse

import requests
from requests.status_codes import _codes  # type:ignore

from tavern._core import exceptions
from tavern._core.dict_util import deep_dict_merge
from tavern._core.pytest.config import TestConfig
from tavern._core.pytest.newhooks import call_hook
from tavern._core.report import attach_yaml
from tavern.response import BaseResponse, indent_err_text

logger: logging.Logger = logging.getLogger(__name__)


class RestResponse(BaseResponse):
    response: requests.Response

    def __init__(
        self, session, name: str, expected, test_block_config: TestConfig
    ) -> None:
        defaults = {"status_code": 200}

        super().__init__(name, deep_dict_merge(defaults, expected), test_block_config)

        def check_code(code: int) -> None:
            if int(code) not in _codes:
                logger.warning("Unexpected status code '%s'", code)

        in_file = self.expected["status_code"]
        try:
            if isinstance(in_file, list):
                for code_ in in_file:
                    check_code(code_)
            else:
                check_code(in_file)
        except TypeError as e:
            raise exceptions.BadSchemaError("Invalid code") from e

    def __str__(self) -> str:
        if self.response:
            return self.response.text.strip()
        else:
            return "<Not run yet>"

    def _verbose_log_response(self, response: requests.Response) -> None:
        """Verbosely log the response object, with query params etc."""

        logger.info("Response: '%s'", response)

        def log_dict_block(block, name):
            if block:
                to_log = name + ":"

                if isinstance(block, list):
                    for v in block:
                        to_log += f"\n  - {v}"
                elif isinstance(block, dict):
                    for k, v in block.items():
                        to_log += f"\n  {k}: {v}"
                else:
                    to_log += f"\n {block}"
                logger.debug(to_log)

        log_dict_block(response.headers, "Headers")

        with contextlib.suppress(ValueError):
            log_dict_block(response.json(), "Body")

        redirect_query_params = self._get_redirect_query_params(response)
        if redirect_query_params:
            parsed_url = urlparse(response.headers["location"])
            to_path = "{}://{}{}".format(*parsed_url)
            logger.debug("Redirect location: %s", to_path)
            log_dict_block(redirect_query_params, "Redirect URL query parameters")

    def _get_redirect_query_params(self, response: requests.Response) -> Dict[str, str]:
        """If there was a redirect header, get any query parameters from it"""

        try:
            redirect_url = response.headers["location"]
        except KeyError as e:
            if "redirect_query_params" in self.expected.get("save", {}):
                self._adderr(
                    "Wanted to save %s, but there was no redirect url in response",
                    self.expected["save"]["redirect_query_params"],
                    e=e,
                )
            redirect_query_params = {}
        else:
            parsed = urlparse(redirect_url)
            qp = parsed.query
            redirect_query_params = {i: j[0] for i, j in parse_qs(qp).items()}

        return redirect_query_params

    def _check_status_code(self, status_code: Union[int, List[int]], body: Any) -> None:
        expected_code = self.expected["status_code"]

        if (isinstance(expected_code, int) and status_code == expected_code) or (
            isinstance(expected_code, list) and (status_code in expected_code)
        ):
            logger.debug(
                "Status code '%s' matched expected '%s'", status_code, expected_code
            )
            return
        elif isinstance(status_code, int) and 400 <= status_code < 500:
            # special case if there was a bad request. This assumes that the
            # response would contain some kind of information as to why this
            # request was rejected.
            self._adderr(
                "Status code was %s, expected %s:\n%s",
                status_code,
                expected_code,
                indent_err_text(json.dumps(body)),
            )
        else:
            self._adderr("Status code was %s, expected %s", status_code, expected_code)

    def verify(self, response: requests.Response) -> dict:
        """Verify response against expected values and returns any values that
        we wanted to save for use in future requests

        There are various ways to 'validate' a block - a specific function, just
        matching values, validating a schema, etc...

        Args:
            response: response object

        Returns:
            Any saved values

        Raises:
            TestFailError: Something went wrong with validating the response
        """
        self._verbose_log_response(response)

        call_hook(
            self.test_block_config,
            "pytest_tavern_beta_after_every_response",
            expected=self.expected,
            response=response,
        )

        self.response = response

        # Get things to use from the response
        try:
            body = response.json()
        except ValueError:
            body = None

        redirect_query_params = self._get_redirect_query_params(response)

        # Run validation on response
        self._check_status_code(response.status_code, body)

        self._validate_block("json", body)
        self._validate_block("headers", response.headers)
        self._validate_block("redirect_query_params", redirect_query_params)

        attach_yaml(
            {
                "status_code": response.status_code,
                "headers": dict(response.headers),
                "body": body,
                "redirect_query_params": redirect_query_params,
            },
            name="rest_response",
        )

        self._maybe_run_validate_functions(response)

        # Get any keys to save
        saved: Dict = {}

        saved.update(self.maybe_get_save_values_from_save_block("json", body))
        saved.update(
            self.maybe_get_save_values_from_save_block("headers", response.headers)
        )
        saved.update(
            self.maybe_get_save_values_from_save_block(
                "redirect_query_params", redirect_query_params
            )
        )

        saved.update(self.maybe_get_save_values_from_ext(response, self.expected))

        # Check cookies
        for cookie in self.expected.get("cookies", []):
            if cookie not in response.cookies:
                self._adderr("No cookie named '%s' in response", cookie)

        if self.errors:
            raise exceptions.TestFailError(
                f"Test '{self.name:s}' failed:\n{self._str_errors():s}",
                failures=self.errors,
            )

        return saved

    def _validate_block(self, blockname: str, block: Mapping) -> None:
        """Validate a block of the response

        Args:
            blockname: which part of the response is being checked
            block: The actual part being checked
        """
        try:
            expected_block = self.expected[blockname]
        except KeyError:
            expected_block = None

        if isinstance(expected_block, dict):
            if expected_block.pop("$ext", None):
                raise exceptions.MisplacedExtBlockException(
                    blockname,
                )

        if blockname == "headers" and expected_block is not None:
            # Special case for headers. These need to be checked in a case
            # insensitive manner
            block = {i.lower(): j for i, j in block.items()}
            expected_block = {i.lower(): j for i, j in expected_block.items()}

        logger.debug("Validating response %s against %s", blockname, expected_block)

        test_strictness = self.test_block_config.strict
        block_strictness = test_strictness.option_for(blockname)
        self.recurse_check_key_match(expected_block, block, blockname, block_strictness)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
import test_services_pb2 as test__services__pb2


class DummyServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Empty = channel.unary_unary(
                '/tavern.tests.v1.DummyService/Empty',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.SimpleTest = channel.unary_unary(
                '/tavern.tests.v1.DummyService/SimpleTest',
                request_serializer=test__services__pb2.DummyRequest.SerializeToString,
                response_deserializer=test__services__pb2.DummyResponse.FromString,
                )


class DummyServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Empty(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SimpleTest(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DummyServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Empty': grpc.unary_unary_rpc_method_handler(
                    servicer.Empty,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'SimpleTest': grpc.unary_unary_rpc_method_handler(
                    servicer.SimpleTest,
                    request_deserializer=test__services__pb2.DummyRequest.FromString,
                    response_serializer=test__services__pb2.DummyResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'tavern.tests.v1.DummyService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class DummyService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Empty(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tavern.tests.v1.DummyService/Empty',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SimpleTest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tavern.tests.v1.DummyService/SimpleTest',
            test__services__pb2.DummyRequest.SerializeToString,
            test__services__pb2.DummyResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import logging
import time
from functools import wraps
from typing import Callable, Mapping, Union

from tavern._core import exceptions
from tavern._core.dict_util import format_keys
from tavern._core.pytest.config import TestConfig

logger: logging.Logger = logging.getLogger(__name__)


def delay(stage: Mapping, when: str, variables: Mapping) -> None:
    """Look for delay_before/delay_after and sleep

    Args:
        stage: test stage
        when: 'before' or 'after'
        variables: Variables to format with
    """

    try:
        length = format_keys(stage[f"delay_{when}"], variables)
    except KeyError:
        pass
    else:
        logger.debug("Delaying %s request for %.2f seconds", when, length)
        time.sleep(length)


def retry(stage: Mapping, test_block_config: TestConfig) -> Callable:
    """Look for retry and try to repeat the stage `retry` times.

    Args:
        stage: test stage
        test_block_config: Configuration for current test
    """

    if r := stage.get("max_retries", None):
        max_retries = maybe_format_max_retries(r, test_block_config)
    else:
        max_retries = 0

    if max_retries == 0:

        def catch_wrapper(fn):
            @wraps(fn)
            def wrapped(*args, **kwargs):
                res = fn(*args, **kwargs)
                logger.debug("Stage '%s' succeeded.", stage["name"])
                return res

            return wrapped

        return catch_wrapper
    else:

        def retry_wrapper(fn):
            @wraps(fn)
            def wrapped(*args, **kwargs):
                i = 0
                res = None
                for i in range(max_retries + 1):
                    try:
                        res = fn(*args, **kwargs)
                    except exceptions.BadSchemaError:
                        raise
                    except exceptions.TavernException as e:
                        if i < max_retries:
                            logger.info(
                                "Stage '%s' failed for %i time. Retrying.",
                                stage["name"],
                                i + 1,
                            )
                            delay(stage, "after", test_block_config.variables)
                        else:
                            logger.error(
                                "Stage '%s' did not succeed in %i retries.",
                                stage["name"],
                                max_retries,
                            )

                            if isinstance(e, exceptions.TestFailError):
                                raise
                            else:
                                raise exceptions.TestFailError(
                                    "Test '{}' failed: stage did not succeed in {} retries.".format(
                                        stage["name"], max_retries
                                    )
                                ) from e
                    else:
                        break

                logger.debug("Stage '%s' succeed after %i retries.", stage["name"], i)
                return res

            return wrapped

        return retry_wrapper


def maybe_format_max_retries(
    max_retries: Union[str, int], test_block_config: TestConfig
) -> int:
    """Possibly handle max_retries validation"""

    # Probably a format variable, or just invalid (in which case it will fail further down)
    max_retries = int(format_keys(max_retries, test_block_config.variables))  # type:ignore

    # Missing type token will mean that max_retries is still a string and will fail here
    # Could auto convert here as well, but keep it consistent and just fail
    if not isinstance(max_retries, int):
        raise exceptions.InvalidRetryException(
            f"Invalid type for max_retries - was {type(max_retries)}"
        )

    if max_retries < 0:
        raise exceptions.InvalidRetryException("max_retries must be greater than 0")

    return max_retries
